import scanpy as sc
from fastmcp.tools.tool import Tool
from fastmcp.exceptions import ToolError
from scmcp_shared.schema.preset import AdataInfo
from scmcp_shared.schema.preset.pp import *
from scmcp_shared.util import (
    filter_args,
    add_op_log,
    forward_request,
    get_ads,
    generate_msg,
)
from scmcp_shared.mcp_base import BaseMCP


class ScanpyPreprocessingMCP(BaseMCP):
    def __init__(
        self,
        include_tools: list = None,
        exclude_tools: list = None,
        AdataInfo: AdataInfo = AdataInfo,
    ):
        """
        Initialize ScanpyPreprocessingMCP with optional tool filtering.

        Args:
            include_tools (list, optional): List of tool names to include. If None, all tools are included.
            exclude_tools (list, optional): List of tool names to exclude. If None, no tools are excluded.
            AdataInfo: The AdataInfo class to use for type annotations.
        """
        super().__init__("ScanpyMCP-PP-Server", include_tools, exclude_tools, AdataInfo)

    def _tool_subset_cells(self):
        def _subset_cells(
            request: SubsetCellParam, adinfo: self.AdataInfo = self.AdataInfo()
        ):
            """filter or subset cells based on total genes expressed counts and numbers. or values in adata.obs[obs_key]"""
            try:
                result = forward_request("subset_cells", request, adinfo)
                if result is not None:
                    return result

                ads = get_ads()
                adata = ads.get_adata(adinfo=adinfo).copy()
                func_kwargs = filter_args(request, sc.pp.filter_cells)
                if func_kwargs:
                    sc.pp.filter_cells(adata, **func_kwargs)
                    add_op_log(adata, sc.pp.filter_cells, func_kwargs, adinfo)
                # Subset based on obs (cells) criteria
                if request.obs_key is not None:
                    if request.obs_key not in adata.obs.columns:
                        raise ValueError(
                            f"Key '{request.obs_key}' not found in adata.obs"
                        )
                    mask = True  # Start with all cells selected
                    if request.obs_value is not None:
                        mask = mask & (adata.obs[request.obs_key] == request.obs_value)
                    if request.obs_min is not None:
                        mask = mask & (adata.obs[request.obs_key] >= request.obs_min)
                    if request.obs_max is not None:
                        mask = mask & (adata.obs[request.obs_key] <= request.obs_max)
                    adata = adata[mask, :]
                    add_op_log(
                        adata,
                        "subset_cells",
                        {
                            "obs_key": request.obs_key,
                            "obs_value": request.obs_value,
                            "obs_min": request.obs_min,
                            "obs_max": request.obs_max,
                        },
                        adinfo,
                    )
                ads.set_adata(adata, adinfo=adinfo)
                return [generate_msg(adinfo, adata, ads)]
            except ToolError as e:
                raise ToolError(e)
            except Exception as e:
                if hasattr(e, "__context__") and e.__context__:
                    raise ToolError(e.__context__)
                else:
                    raise ToolError(e)

        return Tool.from_function(_subset_cells, name="subset_cells", enabled=True)

    def _tool_subset_genes(self):
        def _subset_genes(
            request: SubsetGeneParam, adinfo: self.AdataInfo = self.AdataInfo()
        ):
            """filter or subset genes based on number of cells or counts, or values in adata.var[var_key] or subset highly variable genes"""
            try:
                result = forward_request("pp_subset_genes", request, adinfo)
                if result is not None:
                    return result
                func_kwargs = filter_args(request, sc.pp.filter_genes)
                ads = get_ads()
                adata = ads.get_adata(adinfo=adinfo).copy()
                if request.highly_variable:
                    adata = adata[:, adata.var.highly_variable]
                    add_op_log(adata, "subset_genes", {"hpv": "true"}, adinfo)
                if func_kwargs:
                    sc.pp.filter_genes(adata, **func_kwargs)
                    add_op_log(adata, sc.pp.filter_genes, func_kwargs, adinfo)
                if request.var_key is not None:
                    if request.var_key not in adata.var.columns:
                        raise ValueError(
                            f"Key '{request.var_key}' not found in adata.var"
                        )
                    mask = True  # Start with all genes selected
                    if request.var_min is not None:
                        mask = mask & (adata.var[request.var_key] >= request.var_min)
                    if request.var_max is not None:
                        mask = mask & (adata.var[request.var_key] <= request.var_max)
                    adata = adata[:, mask]
                    add_op_log(
                        adata,
                        "subset_genes",
                        {
                            "var_key": request.var_key,
                            "var_min": request.var_min,
                            "var_max": request.var_max,
                            "hpv": request.highly_variable,
                        },
                        adinfo,
                    )
                ads.set_adata(adata, adinfo=adinfo)
                return [generate_msg(adinfo, adata, ads)]
            except ToolError as e:
                raise ToolError(e)
            except Exception as e:
                if hasattr(e, "__context__") and e.__context__:
                    raise ToolError(e.__context__)
                else:
                    raise ToolError(e)

        return Tool.from_function(_subset_genes, name="subset_genes", enabled=True)

    def _tool_calculate_qc_metrics(self):
        def _calculate_qc_metrics(
            request: CalculateQCMetrics, adinfo: self.AdataInfo = self.AdataInfo()
        ):
            """Calculate quality control metrics(common metrics: total counts, gene number, percentage of counts in ribosomal and mitochondrial) for AnnData."""
            try:
                result = forward_request("pp_calculate_qc_metrics", request, adinfo)
                if result is not None:
                    return result

                func_kwargs = filter_args(request, sc.pp.calculate_qc_metrics)
                ads = get_ads()
                adata = ads.get_adata(adinfo=adinfo)
                if request.qc_vars:
                    for var in request.qc_vars:
                        if var not in adata.var.columns:
                            return f"Cound find {var} in adata.var, consider to use mark_var tool to mark the variable"
                func_kwargs["inplace"] = True
                try:
                    sc.pp.calculate_qc_metrics(adata, **func_kwargs)
                    add_op_log(adata, sc.pp.calculate_qc_metrics, func_kwargs, adinfo)
                except KeyError as e:
                    raise KeyError(
                        f"Cound find {e} in adata.var, consider to use mark_var tool to mark the variable"
                    )
                return [generate_msg(adinfo, adata, ads)]
            except ToolError as e:
                raise ToolError(e)
            except Exception as e:
                if hasattr(e, "__context__") and e.__context__:
                    raise ToolError(e.__context__)
                else:
                    raise ToolError(e)

        return Tool.from_function(
            _calculate_qc_metrics, name="calculate_qc_metrics", enabled=True
        )

    def _tool_log1p(self):
        def _log1p(
            request: Log1PParam = Log1PParam(),
            adinfo: self.AdataInfo = self.AdataInfo(),
        ):
            """Logarithmize the data matrix"""
            try:
                result = forward_request("pp_log1p", request, adinfo)
                if result is not None:
                    return result
                func_kwargs = filter_args(request, sc.pp.log1p)
                ads = get_ads()
                adata = ads.get_adata(adinfo=adinfo).copy()
                try:
                    sc.pp.log1p(adata, **func_kwargs)
                    adata.raw = adata.copy()
                    add_op_log(adata, sc.pp.log1p, func_kwargs, adinfo)
                except Exception as e:
                    raise e
                ads.set_adata(adata, adinfo=adinfo)
                return [generate_msg(adinfo, adata, ads)]
            except ToolError as e:
                raise ToolError(e)
            except Exception as e:
                if hasattr(e, "__context__") and e.__context__:
                    raise ToolError(e.__context__)
                else:
                    raise ToolError(e)

        return Tool.from_function(_log1p, name="log1p", enabled=True)

    def _tool_normalize_total(self):
        def _normalize_total(
            request: NormalizeTotalParam, adinfo: self.AdataInfo = self.AdataInfo()
        ):
            """Normalize counts per cell to the same total count"""
            try:
                result = forward_request("pp_normalize_total", request, adinfo)
                if result is not None:
                    return result
                func_kwargs = filter_args(request, sc.pp.normalize_total)
                ads = get_ads()
                adata = ads.get_adata(adinfo=adinfo).copy()
                sc.pp.normalize_total(adata, **func_kwargs)
                add_op_log(adata, sc.pp.normalize_total, func_kwargs, adinfo)
                ads.set_adata(adata, adinfo=adinfo)
                return [generate_msg(adinfo, adata, ads)]
            except ToolError as e:
                raise ToolError(e)
            except Exception as e:
                if hasattr(e, "__context__") and e.__context__:
                    raise ToolError(e.__context__)
                else:
                    raise ToolError(e)

        return Tool.from_function(
            _normalize_total, name="normalize_total", enabled=True
        )

    def _tool_highly_variable_genes(self):
        def _highly_variable_genes(
            request: HighlyVariableGenesParam, adinfo: self.AdataInfo = self.AdataInfo()
        ):
            """Annotate highly variable genes"""
            try:
                result = forward_request("pp_highly_variable_genes", request, adinfo)
                if result is not None:
                    return result
                try:
                    func_kwargs = filter_args(request, sc.pp.highly_variable_genes)
                    ads = get_ads()
                    adata = ads.get_adata(adinfo=adinfo)
                    sc.pp.highly_variable_genes(adata, **func_kwargs)
                    add_op_log(adata, sc.pp.highly_variable_genes, func_kwargs, adinfo)
                except Exception as e:
                    raise e
                return [generate_msg(adinfo, adata, ads)]
            except ToolError as e:
                raise ToolError(e)
            except Exception as e:
                if hasattr(e, "__context__") and e.__context__:
                    raise ToolError(e.__context__)
                else:
                    raise ToolError(e)

        return Tool.from_function(
            _highly_variable_genes, name="highly_variable_genes", enabled=True
        )

    def _tool_regress_out(self):
        def _regress_out(
            request: RegressOutParam, adinfo: self.AdataInfo = self.AdataInfo()
        ):
            """Regress out (mostly) unwanted sources of variation."""
            try:
                result = forward_request("pp_regress_out", request, adinfo)
                if result is not None:
                    return result
                func_kwargs = filter_args(request, sc.pp.regress_out)
                ads = get_ads()
                adata = ads.get_adata(adinfo=adinfo).copy()
                sc.pp.regress_out(adata, **func_kwargs)
                add_op_log(adata, sc.pp.regress_out, func_kwargs, adinfo)
                ads.set_adata(adata, adinfo=adinfo)
                return [generate_msg(adinfo, adata, ads)]
            except ToolError as e:
                raise ToolError(e)
            except Exception as e:
                if hasattr(e, "__context__") and e.__context__:
                    raise ToolError(e.__context__)
                else:
                    raise ToolError(e)

        return Tool.from_function(_regress_out, name="regress_out", enabled=True)

    def _tool_scale(self):
        def _scale(request: ScaleParam, adinfo: self.AdataInfo = self.AdataInfo()):
            """Scale data to unit variance and zero mean"""
            try:
                result = forward_request("pp_scale", request, adinfo)
                if result is not None:
                    return result
                func_kwargs = filter_args(request, sc.pp.scale)
                ads = get_ads()
                adata = ads.get_adata(adinfo=adinfo).copy()

                sc.pp.scale(adata, **func_kwargs)
                add_op_log(adata, sc.pp.scale, func_kwargs, adinfo)

                ads.set_adata(adata, adinfo=adinfo)
                return [generate_msg(adinfo, adata, ads)]
            except ToolError as e:
                raise ToolError(e)
            except Exception as e:
                if hasattr(e, "__context__") and e.__context__:
                    raise ToolError(e.__context__)
                else:
                    raise ToolError(e)

        return Tool.from_function(_scale, name="scale", enabled=True)

    def _tool_combat(self):
        def _combat(request: CombatParam, adinfo: self.AdataInfo = self.AdataInfo()):
            """ComBat function for batch effect correction"""
            try:
                result = forward_request("pp_combat", request, adinfo)
                if result is not None:
                    return result
                func_kwargs = filter_args(request, sc.pp.combat)
                ads = get_ads()
                adata = ads.get_adata(adinfo=adinfo).copy()

                sc.pp.combat(adata, **func_kwargs)
                add_op_log(adata, sc.pp.combat, func_kwargs, adinfo)

                ads.set_adata(adata, adinfo=adinfo)
                return [generate_msg(adinfo, adata, ads)]
            except ToolError as e:
                raise ToolError(e)
            except Exception as e:
                if hasattr(e, "__context__") and e.__context__:
                    raise ToolError(e.__context__)
                else:
                    raise ToolError(e)

        return Tool.from_function(_combat, name="combat", enabled=True)

    def _tool_scrublet(self):
        def _scrublet(
            request: ScrubletParam, adinfo: self.AdataInfo = self.AdataInfo()
        ):
            """Predict doublets using Scrublet"""
            try:
                result = forward_request("pp_scrublet", request, adinfo)
                if result is not None:
                    return result
                func_kwargs = filter_args(request, sc.pp.scrublet)
                ads = get_ads()
                adata = ads.get_adata(adinfo=adinfo)
                sc.pp.scrublet(adata, **func_kwargs)
                add_op_log(adata, sc.pp.scrublet, func_kwargs, adinfo)
                return [generate_msg(adinfo, adata, ads)]
            except ToolError as e:
                raise ToolError(e)
            except Exception as e:
                if hasattr(e, "__context__") and e.__context__:
                    raise ToolError(e.__context__)
                else:
                    raise ToolError(e)

        return Tool.from_function(_scrublet, name="scrublet", enabled=True)

    def _tool_neighbors(self):
        def _neighbors(
            request: NeighborsParam, adinfo: self.AdataInfo = self.AdataInfo()
        ):
            """Compute nearest neighbors distance matrix and neighborhood graph"""
            try:
                result = forward_request("pp_neighbors", request, adinfo)
                if result is not None:
                    return result
                func_kwargs = filter_args(request, sc.pp.neighbors)
                ads = get_ads()
                adata = ads.get_adata(adinfo=adinfo)
                sc.pp.neighbors(adata, **func_kwargs)
                add_op_log(adata, sc.pp.neighbors, func_kwargs, adinfo)
                return [generate_msg(adinfo, adata, ads)]
            except ToolError as e:
                raise ToolError(e)
            except Exception as e:
                if hasattr(e, "__context__") and e.__context__:
                    raise ToolError(e.__context__)
                else:
                    raise ToolError(e)

        return Tool.from_function(_neighbors, name="neighbors", enabled=True)



from pydantic import Field, field_validator, BaseModel

from typing import Optional, Union, List, Dict, Any
from typing import Literal
import numpy as np


class FilterCellsParam(BaseModel):
    """Input schema for the filter_cells preprocessing tool."""

    adata: str = Field(..., description="The AnnData object variable name.")
    min_counts: Optional[int] = Field(
        default=None,
        description="Minimum number of counts required for a cell to pass filtering.",
    )

    min_genes: Optional[int] = Field(
        default=None,
        description="Minimum number of genes expressed required for a cell to pass filtering.",
    )

    max_counts: Optional[int] = Field(
        default=None,
        description="Maximum number of counts required for a cell to pass filtering.",
    )

    max_genes: Optional[int] = Field(
        default=None,
        description="Maximum number of genes expressed required for a cell to pass filtering.",
    )

    @field_validator("min_counts", "min_genes", "max_counts", "max_genes")
    def validate_positive_integers(cls, v: Optional[int]) -> Optional[int]:
        if v is not None and v <= 0:
            raise ValueError("must be positive_integers")
        return v


class FilterGenesParam(BaseModel):
    """Input schema for the filter_genes preprocessing tool."""

    adata: str = Field(..., description="The AnnData object variable name.")
    min_counts: Optional[int] = Field(
        default=None,
        description="Minimum number of counts required for a gene to pass filtering.",
    )

    min_cells: Optional[int] = Field(
        default=None,
        description="Minimum number of cells expressed required for a gene to pass filtering.",
    )

    max_counts: Optional[int] = Field(
        default=None,
        description="Maximum number of counts required for a gene to pass filtering.",
    )

    max_cells: Optional[int] = Field(
        default=None,
        description="Maximum number of cells expressed required for a gene to pass filtering.",
    )

    @field_validator("min_counts", "min_cells", "max_counts", "max_cells")
    def validate_positive_integers(cls, v: Optional[int]) -> Optional[int]:
        if v is not None and v <= 0:
            raise ValueError("must be positive_integers")
        return v


class SubsetCellParam(BaseModel):
    """Input schema for subsetting AnnData objects based on various criteria."""

    adata: str = Field(..., description="The AnnData object variable name.")
    obs_key: Optional[str] = Field(
        default=None,
        description="Key in adata.obs to use for subsetting observations/cells.",
    )
    obs_min: Optional[float] = Field(
        default=None,
        description="Minimum value for the obs_key to include in the subset.",
    )
    obs_max: Optional[float] = Field(
        default=None,
        description="Maximum value for the obs_key to include in the subset.",
    )
    obs_value: Optional[Any] = Field(
        default=None,
        description="Exact value for the obs_key to include in the subset (adata.obs[obs_key] == obs_value).",
    )
    min_counts: Optional[int] = Field(
        default=None,
        description="Minimum number of counts required for a cell to pass filtering.",
    )
    min_genes: Optional[int] = Field(
        default=None,
        description="Minimum number of genes expressed required for a cell to pass filtering.",
    )
    max_counts: Optional[int] = Field(
        default=None,
        description="Maximum number of counts required for a cell to pass filtering.",
    )
    max_genes: Optional[int] = Field(
        default=None,
        description="Maximum number of genes expressed required for a cell to pass filtering.",
    )


class SubsetGeneParam(BaseModel):
    """Input schema for subsetting AnnData objects based on various criteria."""

    adata: str = Field(..., description="The AnnData object variable name.")
    min_counts: Optional[int] = Field(
        default=None,
        description="Minimum number of counts required for a gene to pass filtering.",
    )
    min_cells: Optional[int] = Field(
        default=None,
        description="Minimum number of cells expressed required for a gene to pass filtering.",
    )
    max_counts: Optional[int] = Field(
        default=None,
        description="Maximum number of counts required for a gene to pass filtering.",
    )
    max_cells: Optional[int] = Field(
        default=None,
        description="Maximum number of cells expressed required for a gene to pass filtering.",
    )
    var_key: Optional[str] = Field(
        default=None,
        description="Key in adata.var to use for subsetting variables/genes.",
    )
    var_min: Optional[float] = Field(
        default=None,
        description="Minimum value for the var_key to include in the subset.",
    )
    var_max: Optional[float] = Field(
        default=None,
        description="Maximum value for the var_key to include in the subset.",
    )
    highly_variable: Optional[bool] = Field(
        default=False,
        description="If True, subset to highly variable genes. Requires 'highly_variable' column in adata.var.",
    )


class CalculateQCMetrics(BaseModel):
    """Input schema for the calculate_qc_metrics preprocessing tool."""

    adata: str = Field(..., description="The AnnData object variable name.")

    expr_type: str = Field(default="counts", description="Name of kind of values in X.")

    var_type: str = Field(
        default="genes", description="The kind of thing the variables are."
    )

    qc_vars: Optional[Union[List[str], str]] = Field(
        default=[],
        description=(
            "Keys for boolean columns of .var which identify variables you could want to control for "
            "mark_var tool should be called frist when you want to calculate mt, ribo, hb, and check tool output for var columns"
        ),
    )

    percent_top: Optional[List[int]] = Field(
        default=[50, 100, 200, 500],
        description="List of ranks (where genes are ranked by expression) at which the cumulative proportion of expression will be reported as a percentage.",
    )

    layer: Optional[str] = Field(
        default=None,
        description="If provided, use adata.layers[layer] for expression values instead of adata.X",
    )

    use_raw: bool = Field(
        default=False,
        description="If True, use adata.raw.X for expression values instead of adata.X",
    )
    log1p: bool = Field(
        default=True,
        description="Set to False to skip computing log1p transformed annotations.",
    )

    @field_validator("percent_top")
    def validate_percent_top(cls, v: Optional[List[int]]) -> Optional[List[int]]:
        """验证 percent_top 中的值为正整数"""
        if v is not None:
            for rank in v:
                if not isinstance(rank, int) or rank <= 0:
                    raise ValueError("percent_top 中的所有值必须是正整数")
        return v


class Log1PParam(BaseModel):
    """Input schema for the log1p preprocessing tool."""

    adata: str = Field(..., description="The AnnData object variable name.")
    base: Optional[Union[int, float]] = Field(
        default=None,
        description="Base of the logarithm. Natural logarithm is used by default.",
    )

    chunked: Optional[bool] = Field(
        default=None,
        description="Process the data matrix in chunks, which will save memory.",
    )

    chunk_size: Optional[int] = Field(
        default=None,
        description="Number of observations in the chunks to process the data in.",
    )

    layer: Optional[str] = Field(
        default=None, description="Entry of layers to transform."
    )

    obsm: Optional[str] = Field(default=None, description="Entry of obsm to transform.")

    @field_validator("chunk_size")
    def validate_chunk_size(cls, v: Optional[int]) -> Optional[int]:
        """Validate chunk_size is positive integer"""
        if v is not None and v <= 0:
            raise ValueError("chunk_size must be a positive integer")
        return v


class HighlyVariableGenesParam(BaseModel):
    """Input schema for the highly_variable_genes preprocessing tool."""

    adata: str = Field(..., description="The AnnData object variable name.")
    layer: Optional[str] = Field(
        default=None,
        description="If provided, use adata.layers[layer] for expression values.",
    )

    n_top_genes: Optional[int] = Field(
        default=None,
        description="Number of highly-variable genes to keep. Mandatory if `flavor='seurat_v3'",
    )

    min_disp: Optional[float] = Field(
        default=0.5, description="Minimum dispersion cutoff for gene selection."
    )

    max_disp: Optional[float] = Field(
        default=np.inf, description="Maximum dispersion cutoff for gene selection."
    )
    min_mean: Optional[float] = Field(
        default=0.0125, description="Minimum mean expression cutoff for gene selection."
    )
    max_mean: Optional[float] = Field(
        default=3, description="Maximum mean expression cutoff for gene selection."
    )
    span: Optional[float] = Field(
        default=0.3,
        description="Fraction of data used for loess model fit in seurat_v3.",
        gt=0,
        lt=1,
    )
    n_bins: Optional[int] = Field(
        default=20, description="Number of bins for mean expression binning.", gt=0
    )
    flavor: Optional[
        Literal["seurat", "cell_ranger", "seurat_v3", "seurat_v3_paper"]
    ] = Field(
        default="seurat", description="Method for identifying highly variable genes."
    )
    subset: Optional[bool] = Field(
        default=False, description="Inplace subset to highly-variable genes if True."
    )
    batch_key: Optional[str] = Field(
        default=None, description="Key in adata.obs for batch information."
    )

    check_values: Optional[bool] = Field(
        default=True, description="Check if counts are integers for seurat_v3 flavor."
    )

    @field_validator("n_top_genes", "n_bins")
    def validate_positive_integers(cls, v: Optional[int]) -> Optional[int]:
        """Validate positive integers"""
        if v is not None and v <= 0:
            raise ValueError("must be a positive integer")
        return v

    @field_validator("span")
    def validate_span(cls, v: float) -> float:
        """Validate span is between 0 and 1"""
        if v <= 0 or v >= 1:
            raise ValueError("span must be between 0 and 1")
        return v


class RegressOutParam(BaseModel):
    """Input schema for the regress_out preprocessing tool."""

    adata: str = Field(..., description="The AnnData object variable name.")
    keys: Union[str, List[str]] = Field(
        description="Keys for observation annotation on which to regress on."
    )
    layer: Optional[str] = Field(
        default=None, description="If provided, which element of layers to regress on."
    )
    n_jobs: Optional[int] = Field(
        default=None, description="Number of jobs for parallel computation.", gt=0
    )

    @field_validator("n_jobs")
    def validate_n_jobs(cls, v: Optional[int]) -> Optional[int]:
        """Validate n_jobs is positive integer"""
        if v is not None and v <= 0:
            raise ValueError("n_jobs must be a positive integer")
        return v

    @field_validator("keys")
    def validate_keys(cls, v: Union[str, List[str]]) -> Union[str, List[str]]:
        """Ensure keys is either a string or list of strings"""
        if isinstance(v, str):
            return v
        elif isinstance(v, list) and all(isinstance(item, str) for item in v):
            return v
        raise ValueError("keys must be a string or list of strings")


class ScaleParam(BaseModel):
    """Input schema for the scale preprocessing tool."""

    adata: str = Field(..., description="The AnnData object variable name.")
    zero_center: bool = Field(
        default=True,
        description="If False, omit zero-centering variables to handle sparse input efficiently.",
    )

    max_value: Optional[float] = Field(
        default=None,
        description="Clip (truncate) to this value after scaling. If None, do not clip.",
    )

    layer: Optional[str] = Field(
        default=None, description="If provided, which element of layers to scale."
    )

    obsm: Optional[str] = Field(
        default=None, description="If provided, which element of obsm to scale."
    )

    mask_obs: Optional[Union[str, bool]] = Field(
        default=None,
        description="Boolean mask or string referring to obs column for subsetting observations.",
    )

    @field_validator("max_value")
    def validate_max_value(cls, v: Optional[float]) -> Optional[float]:
        """Validate max_value is positive if provided"""
        if v is not None and v <= 0:
            raise ValueError("max_value must be positive if provided")
        return v


class CombatParam(BaseModel):
    """Input schema for the combat batch effect correction tool."""

    adata: str = Field(..., description="The AnnData object variable name.")
    key: str = Field(
        default="batch",
        description="Key to a categorical annotation from adata.obs that will be used for batch effect removal.",
    )

    covariates: Optional[List[str]] = Field(
        default=None,
        description="Additional covariates besides the batch variable such as adjustment variables or biological condition.",
    )

    @field_validator("key")
    def validate_key(cls, v: str) -> str:
        """Validate key is not empty"""
        if not v.strip():
            raise ValueError("key cannot be empty")
        return v

    @field_validator("covariates")
    def validate_covariates(cls, v: Optional[List[str]]) -> Optional[List[str]]:
        """Validate covariates are non-empty strings if provided"""
        if v is not None:
            if not all(isinstance(item, str) and item.strip() for item in v):
                raise ValueError("covariates must be non-empty strings")
        return v


class ScrubletParam(BaseModel):
    """Input schema for the scrublet doublet prediction tool."""

    adata: str = Field(..., description="The AnnData object variable name.")
    adata_sim: Optional[str] = Field(
        default=None,
        description="Optional path to AnnData object with simulated doublets.",
    )

    batch_key: Optional[str] = Field(
        default=None, description="Key in adata.obs for batch information."
    )

    sim_doublet_ratio: float = Field(
        default=2.0,
        description="Number of doublets to simulate relative to observed transcriptomes.",
        gt=0,
    )

    expected_doublet_rate: float = Field(
        default=0.05,
        description="Estimated doublet rate for the experiment.",
        ge=0,
        le=1,
    )

    stdev_doublet_rate: float = Field(
        default=0.02,
        description="Uncertainty in the expected doublet rate.",
        ge=0,
        le=1,
    )

    synthetic_doublet_umi_subsampling: float = Field(
        default=1.0,
        description="Rate for sampling UMIs when creating synthetic doublets.",
        gt=0,
        le=1,
    )

    knn_dist_metric: str = Field(
        default="euclidean",
        description="Distance metric used when finding nearest neighbors.",
    )

    normalize_variance: bool = Field(
        default=True,
        description="Normalize data such that each gene has variance of 1.",
    )

    log_transform: bool = Field(
        default=False, description="Whether to log-transform the data prior to PCA."
    )

    mean_center: bool = Field(
        default=True, description="Center data such that each gene has mean of 0."
    )

    n_prin_comps: int = Field(
        default=30,
        description="Number of principal components used for embedding.",
        gt=0,
    )

    use_approx_neighbors: Optional[bool] = Field(
        default=None, description="Use approximate nearest neighbor method (annoy)."
    )

    get_doublet_neighbor_parents: bool = Field(
        default=False,
        description="Return parent transcriptomes that generated doublet neighbors.",
    )

    n_neighbors: Optional[int] = Field(
        default=None,
        description="Number of neighbors used to construct KNN graph.",
        gt=0,
    )

    threshold: Optional[float] = Field(
        default=None,
        description="Doublet score threshold for calling a transcriptome a doublet.",
        ge=0,
        le=1,
    )

    @field_validator(
        "sim_doublet_ratio",
        "expected_doublet_rate",
        "stdev_doublet_rate",
        "synthetic_doublet_umi_subsampling",
        "n_prin_comps",
        "n_neighbors",
    )
    def validate_positive_numbers(
        cls, v: Optional[Union[int, float]]
    ) -> Optional[Union[int, float]]:
        """Validate positive numbers where applicable"""
        if v is not None and v <= 0:
            raise ValueError("must be a positive number")
        return v

    @field_validator("knn_dist_metric")
    def validate_knn_dist_metric(cls, v: str) -> str:
        """Validate distance metric is supported"""
        valid_metrics = ["euclidean", "manhattan", "cosine", "correlation"]
        if v.lower() not in valid_metrics:
            raise ValueError(f"knn_dist_metric must be one of {valid_metrics}")
        return v.lower()


class NeighborsParam(BaseModel):
    """Input schema for the neighbors graph construction tool."""

    adata: str = Field(..., description="The AnnData object variable name.")
    n_neighbors: int = Field(
        default=15,
        description="Size of local neighborhood used for manifold approximation.",
        gt=1,
        le=100,
    )

    n_pcs: Optional[int] = Field(
        default=None,
        description="Number of PCs to use. If None, automatically determined.",
        ge=0,
    )

    use_rep: Optional[str] = Field(
        default=None, description="Key for .obsm to use as representation."
    )

    knn: bool = Field(
        default=True,
        description="Whether to use hard threshold for neighbor restriction.",
    )

    method: Literal["umap", "gauss"] = Field(
        default="umap",
        description="Method for computing connectivities ('umap' or 'gauss').",
    )

    transformer: Optional[str] = Field(
        default=None,
        description="Approximate kNN search implementation ('pynndescent' or 'rapids').",
    )

    metric: str = Field(default="euclidean", description="Distance metric to use.")

    metric_kwds: Dict[str, Any] = Field(
        default_factory=dict, description="Options for the distance metric."
    )

    random_state: int = Field(default=0, description="Random seed for reproducibility.")

    key_added: Optional[str] = Field(
        default=None, description="Key prefix for storing neighbor results."
    )

    @field_validator("n_neighbors", "n_pcs")
    def validate_positive_integers(cls, v: Optional[int]) -> Optional[int]:
        """Validate positive integers where applicable"""
        if v is not None and v <= 0:
            raise ValueError("must be a positive integer")
        return v

    @field_validator("method")
    def validate_method(cls, v: str) -> str:
        """Validate method is supported"""
        if v not in ["umap", "gauss"]:
            raise ValueError("method must be either 'umap' or 'gauss'")
        return v

    @field_validator("transformer")
    def validate_transformer(cls, v: Optional[str]) -> Optional[str]:
        """Validate transformer option is supported"""
        if v is not None and v not in ["pynndescent", "rapids"]:
            raise ValueError("transformer must be either 'pynndescent' or 'rapids'")
        return v


class NormalizeTotalParam(BaseModel):
    """Input schema for the normalize_total preprocessing tool."""

    adata: str = Field(..., description="The AnnData object variable name.")
    target_sum: Optional[float] = Field(
        default=None,
        description="If None, after normalization, each cell has a total count equal to the median of total counts before normalization. If a number is provided, each cell will have this total count after normalization.",
    )

    exclude_highly_expressed: bool = Field(
        default=False,
        description="Exclude highly expressed genes for the computation of the normalization factor for each cell.",
    )

    max_fraction: float = Field(
        default=0.05,
        description="If exclude_highly_expressed=True, consider cells as highly expressed that have more counts than max_fraction of the original total counts in at least one cell.",
        gt=0,
        le=1,
    )

    key_added: Optional[str] = Field(
        default=None,
        description="Name of the field in adata.obs where the normalization factor is stored.",
    )

    layer: Optional[str] = Field(
        default=None,
        description="Layer to normalize instead of X. If None, X is normalized.",
    )

    layers: Optional[Union[Literal["all"], List[str]]] = Field(
        default=None,
        description="List of layers to normalize. If 'all', normalize all layers.",
    )

    layer_norm: Optional[str] = Field(
        default=None, description="Specifies how to normalize layers."
    )

    inplace: bool = Field(
        default=True,
        description="Whether to update adata or return dictionary with normalized copies.",
    )

    @field_validator("target_sum")
    def validate_target_sum(cls, v: Optional[float]) -> Optional[float]:
        """Validate target_sum is positive if provided"""
        if v is not None and v <= 0:
            raise ValueError("target_sum must be positive")
        return v

    @field_validator("max_fraction")
    def validate_max_fraction(cls, v: float) -> float:
        """Validate max_fraction is between 0 and 1"""
        if v <= 0 or v > 1:
            raise ValueError("max_fraction must be between 0 and 1")
        return v


class BBKNNParam(BaseModel):
    """Input schema for the bbknn (batch balanced kNN) preprocessing tool."""

    adata: str = Field(..., description="The AnnData object variable name.")

    batch_key: str = Field(
        default="batch",
        description="adata.obs column name discriminating between your batches.",
    )

    use_rep: str = Field(
        default="X_pca",
        description="The dimensionality reduction in .obsm to use for neighbour detection. Defaults to PCA.",
    )

    approx: bool = Field(
        default=True,
        description="If True, use approximate neighbour finding - annoy or PyNNDescent. This results in a quicker run time for large datasets while also potentially increasing the degree of batch correction.",
    )

    use_annoy: bool = Field(
        default=True,
        description="Only used when approx=True. If True, will use annoy for neighbour finding. If False, will use pyNNDescent instead.",
    )

    metric: str = Field(
        default="euclidean",
        description="What distance metric to use. The options depend on the choice of neighbour algorithm.",
    )

    neighbors_within_batch: int = Field(
        default=3,
        description="How many top neighbours to report for each batch; total number of neighbours in the initial k-nearest-neighbours computation will be this number times the number of batches.",
        gt=0,
    )

    n_pcs: int = Field(
        default=50,
        description="How many dimensions (in case of PCA, principal components) to use in the analysis.",
        gt=0,
    )

    trim: Optional[int] = Field(
        default=None,
        description="Trim the neighbours of each cell to these many top connectivities. May help with population independence and improve the tidiness of clustering. If None, sets the parameter value automatically to 10 times neighbors_within_batch times the number of batches. Set to 0 to skip.",
        ge=0,
    )

    annoy_n_trees: int = Field(
        default=10,
        description="Only used with annoy neighbour identification. The number of trees to construct in the annoy forest. More trees give higher precision when querying, at the cost of increased run time and resource intensity.",
        gt=0,
    )

    pynndescent_n_neighbors: int = Field(
        default=30,
        description="Only used with pyNNDescent neighbour identification. The number of neighbours to include in the approximate neighbour graph. More neighbours give higher precision when querying, at the cost of increased run time and resource intensity.",
        gt=0,
    )

    pynndescent_random_state: int = Field(
        default=0,
        description="Only used with pyNNDescent neighbour identification. The RNG seed to use when creating the graph.",
    )

    use_faiss: bool = Field(
        default=True,
        description="If approx=False and the metric is 'euclidean', use the faiss package to compute nearest neighbours if installed. This improves performance at a minor cost to numerical precision as faiss operates on float32.",
    )

    set_op_mix_ratio: float = Field(
        default=1.0,
        description="UMAP connectivity computation parameter, float between 0 and 1, controlling the blend between a connectivity matrix formed exclusively from mutual nearest neighbour pairs (0) and a union of all observed neighbour relationships with the mutual pairs emphasised (1).",
        ge=0.0,
        le=1.0,
    )

    local_connectivity: int = Field(
        default=1,
        description="UMAP connectivity computation parameter, how many nearest neighbors of each cell are assumed to be fully connected (and given a connectivity value of 1).",
        gt=0,
    )

    @field_validator(
        "neighbors_within_batch",
        "n_pcs",
        "annoy_n_trees",
        "pynndescent_n_neighbors",
        "local_connectivity",
    )
    def validate_positive_integers(cls, v: int) -> int:
        """Validate positive integers"""
        if v <= 0:
            raise ValueError("must be a positive integer")
        return v

    @field_validator("trim")
    def validate_trim(cls, v: Optional[int]) -> Optional[int]:
        """Validate trim is non-negative if provided"""
        if v is not None and v < 0:
            raise ValueError("trim must be non-negative")
        return v

    @field_validator("set_op_mix_ratio")
    def validate_set_op_mix_ratio(cls, v: float) -> float:
        """Validate set_op_mix_ratio is between 0 and 1"""
        if v < 0 or v > 1:
            raise ValueError("set_op_mix_ratio must be between 0 and 1")
        return v

    @field_validator("metric")
    def validate_metric(cls, v: str) -> str:
        """Validate metric is supported"""
        valid_metrics = [
            "euclidean",
            "l2",
            "sqeuclidean",
            "manhattan",
            "taxicab",
            "l1",
            "chebyshev",
            "linfinity",
            "linfty",
            "linf",
            "minkowski",
            "seuclidean",
            "standardised_euclidean",
            "wminkowski",
            "angular",
            "hamming",
        ]
        if v.lower() not in valid_metrics:
            raise ValueError(f"metric must be one of {valid_metrics}")
        return v.lower()


class HarmonyIntegrateParam(BaseModel):
    """Input schema for the harmony_integrate preprocessing tool."""

    adata: str = Field(..., description="The AnnData object variable name.")

    key: Union[str, List[str]] = Field(
        description="The name of the column in adata.obs that differentiates among experiments/batches. To integrate over two or more covariates, you can pass multiple column names as a list."
    )

    basis: str = Field(
        default="X_pca",
        description="The name of the field in adata.obsm where the PCA table is stored. Defaults to 'X_pca', which is the default for sc.pp.pca().",
    )

    adjusted_basis: str = Field(
        default="X_pca_harmony",
        description="The name of the field in adata.obsm where the adjusted PCA table will be stored after running this function. Defaults to X_pca_harmony.",
    )

    theta: float = Field(
        default=2.0,
        description="Diversity clustering penalty parameter. Theta = 0 does not encourage any diversity. Larger values of theta result in more diverse clusters.",
        gt=0,
    )

    lambda_: float = Field(
        default=1.0,
        description="Ridge regression penalty parameter. Lambda = 0 gives no regularization. Larger values of lambda result in more regularization.",
        ge=0,
    )

    sigma: float = Field(
        default=0.1,
        description="Width of soft kmeans clusters. Sigma scales the distance from a cell to cluster centroids. Larger values of sigma result in cells assigned to more clusters. Each cell is assigned to clusters with probability proportional to exp(-distance^2 / sigma^2).",
        gt=0,
    )

    nclust: Optional[int] = Field(
        default=None,
        description="Number of clusters in Harmony. If None, estimated automatically. If provided, this overrides the theta parameter.",
        gt=0,
    )

    tau: float = Field(
        default=0.0,
        description="Protection against overclustering small datasets with rare cell types. tau is the expected number of cells per cluster.",
        ge=0,
    )

    block_size: float = Field(
        default=0.05,
        description="What proportion of cells to update during clustering. Between 0 to 1, e.g. 0.05 updates 5% of cells per iteration.",
        gt=0,
        le=1,
    )

    max_iter_harmony: int = Field(
        default=10,
        description="Maximum number of rounds to run Harmony. One round of Harmony involves one clustering and one correction step.",
        gt=0,
    )

    max_iter_kmeans: int = Field(
        default=20,
        description="Maximum number of rounds to run clustering at each round of Harmony. If at least k < nclust clusters contain 1 or fewer cells, then stop early.",
        gt=0,
    )

    epsilon_cluster: float = Field(
        default=1e-5,
        description="Convergence tolerance for clustering round of Harmony. Set to -Inf to never stop early.",
        gt=0,
    )

    epsilon_harmony: float = Field(
        default=1e-4,
        description="Convergence tolerance for Harmony. Set to -Inf to never stop early.",
        gt=0,
    )

    random_state: int = Field(default=0, description="Random seed for reproducibility.")

    verbose: bool = Field(
        default=True, description="Whether to print progress messages."
    )

    @field_validator("key")
    def validate_key(cls, v: Union[str, List[str]]) -> Union[str, List[str]]:
        """Ensure key is either a string or list of strings"""
        if isinstance(v, str):
            if not v.strip():
                raise ValueError("key cannot be empty")
            return v
        elif isinstance(v, list) and all(
            isinstance(item, str) and item.strip() for item in v
        ):
            return v
        raise ValueError("key must be a non-empty string or list of non-empty strings")

    @field_validator("basis", "adjusted_basis")
    def validate_basis_names(cls, v: str) -> str:
        """Validate basis names are not empty"""
        if not v.strip():
            raise ValueError("basis and adjusted_basis cannot be empty")
        return v

    @field_validator(
        "theta",
        "lambda_",
        "sigma",
        "tau",
        "block_size",
        "epsilon_cluster",
        "epsilon_harmony",
    )
    def validate_positive_floats(cls, v: float) -> float:
        """Validate positive floats"""
        if v < 0:
            raise ValueError("must be non-negative")
        return v

    @field_validator("max_iter_harmony", "max_iter_kmeans")
    def validate_positive_integers(cls, v: int) -> int:
        """Validate positive integers"""
        if v <= 0:
            raise ValueError("must be a positive integer")
        return v

    @field_validator("nclust")
    def validate_nclust(cls, v: Optional[int]) -> Optional[int]:
        """Validate nclust is positive if provided"""
        if v is not None and v <= 0:
            raise ValueError("nclust must be positive if provided")
        return v



import pytest
from scmcp_shared.backend import AdataManager
from scmcp_shared.mcp_base import BaseMCPManager
from scmcp_shared.server.preset import ScanpyIOMCP
from scmcp_shared.server.preset import ScanpyPreprocessingMCP
from scmcp_shared.server.preset import ScanpyToolsMCP
from scmcp_shared.server.preset import ScanpyPlottingMCP
from scmcp_shared.server.preset import ScanpyUtilMCP
from scmcp_shared.server.auto import auto_mcp


class ScanpyMCPManager(BaseMCPManager):
    """Manager class for Scanpy MCP modules."""

    def init_mcp(self):
        """Initialize available Scanpy MCP modules."""
        self.available_modules = {
            "io": ScanpyIOMCP().mcp,
            "pp": ScanpyPreprocessingMCP().mcp,
            "tl": ScanpyToolsMCP().mcp,
            "pl": ScanpyPlottingMCP().mcp,
            "ul": ScanpyUtilMCP().mcp,
            "auto": auto_mcp,
        }


@pytest.fixture
def mcp():
    return ScanpyMCPManager("scmcp", backend=AdataManager).mcp



from .schema.tool import ToolList
import os


from agno.agent import Agent
from agno.models.openai import OpenAILike
from scmcp_shared.kb import load_kb

model = OpenAILike(
    id=os.getenv("MODEL"),
    base_url=os.getenv("BASE_URL"),
    api_key=os.getenv("API_KEY"),
)


def rag_agent(task, software=None):
    knowledge_base = load_kb(software=software)
    agent = Agent(
        model=model,
        knowledge=knowledge_base,
        show_tool_calls=True,
        search_knowledge=True,
    )
    query = f"""
    <task>
    {task}
    </task>
    查询知识库，给出一个用于解决任务的代码示例。返回结果格式为：
    <code_example>
        [code_example]
    </code_example>
    """
    rep = agent.run(query)
    return rep.content


def select_tool(query):
    agent = Agent(
        model=model,
        response_model=ToolList,
        use_json_mode=True,
        instructions="""
        you are a bioinformatician, you are given a task and a list of tools, you need to select the most directly relevant tools to use to solve the task
        """,
    )
    rep = agent.run(query)
    return rep.content



import logging
import sys

from .util import get_env


def setup_logger(name="sc-mcp-server", log_file=None):
    logger = logging.getLogger(name)
    logger.setLevel(logging.INFO)
    if logger.handlers:
        return logger

    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s", "%Y-%m-%d %H:%M:%S"
    )
    if log_file is None:
        log_file = get_env("LOG_FILE")
    if log_file:
        log_handler = logging.FileHandler(log_file)
        log_handler.setFormatter(formatter)
        logger.addHandler(log_handler)

        logger.info(f"logging output: {log_file}")
    else:
        log_handler = logging.StreamHandler(sys.stdout)
        log_handler.setFormatter(formatter)
        logger.addHandler(log_handler)
        logger.info("loggin file output: stdout")
    return logger



from collections.abc import Iterable
from abcoder.backend import NotebookManager

__all__ = ["AdataManager", "NotebookManager"]


class AdataManager:
    def __init__(self, add_adtypes=None):
        self.adata_dic = {"exp": {}, "activity": {}, "cnv": {}, "splicing": {}}
        if isinstance(add_adtypes, str):
            self.adata_dic[add_adtypes] = {}
        elif isinstance(add_adtypes, Iterable):
            self.adata_dic.update({adtype: {} for adtype in add_adtypes})
        self.active_id = None
        self.metadatWa = {}
        self.cr_kernel = {}
        self.cr_estimator = {}

    def get_adata(self, sampleid=None, adtype="exp", adinfo=None):
        if adinfo is not None:
            kwargs = adinfo.model_dump()
            sampleid = kwargs.get("sampleid", None)
            adtype = kwargs.get("adtype", "exp")
        try:
            if self.active_id is None:
                return None
            sampleid = sampleid or self.active_id
            return self.adata_dic[adtype][sampleid]
        except KeyError as e:
            raise KeyError(
                f"Key {e} not found in adata_dic[{adtype}].Please check the sampleid or adtype."
            )
        except Exception as e:
            raise Exception(f"fuck {e} {type(e)}")

    def set_adata(self, adata, sampleid=None, sdtype="exp", adinfo=None):
        if adinfo is not None:
            kwargs = adinfo.model_dump()
            sampleid = kwargs.get("sampleid", None)
            sdtype = kwargs.get("adtype", "exp")
        sampleid = sampleid or self.active_id
        if sdtype not in self.adata_dic:
            self.adata_dic[sdtype] = {}
        self.adata_dic[sdtype][sampleid] = adata



import inspect
from fastmcp import FastMCP
from .schema import AdataInfo
from .util import filter_tools
from collections.abc import AsyncIterator
from contextlib import asynccontextmanager
import asyncio
from typing import Optional, List, Any
from abcoder.backend import NotebookManager


# class BaseMCP:
#     """Base class for all Scanpy MCP classes."""

#     def __init__(self, name: str, instructions=None, lifespan=None, include_tools: list = None, exclude_tools: list = None, AdataInfo = AdataInfo):
#         """
#         Initialize BaseMCP with optional tool filtering.

#         Args:
#             name (str): Name of the MCP server
#             include_tools (list, optional): List of tool names to include. If None, all tools are included.
#             exclude_tools (list, optional): List of tool names to exclude. If None, no tools are excluded.
#             AdataInfo: The AdataInfo class to use for type annotations.
#         """
#         self.mcp = FastMCP(name, instructions=instructions, lifespan=lifespan)
#         self.include_tools = include_tools
#         self.exclude_tools = exclude_tools
#         self.AdataInfo = AdataInfo
#         self._register_tools()

#     def _register_tools(self):
#         """Register all tool methods with the FastMCP instance based on include/exclude filters"""
#         # Get all methods of the class
#         methods = inspect.getmembers(self, predicate=inspect.ismethod)

#         # Filter methods that start with _tool_
#         tool_methods = [tl_method() for name, tl_method in methods if name.startswith('_tool_')]

#         # Filter tools based on include/exclude lists
#         if self.include_tools is not None:
#             tool_methods = [tl for tl in tool_methods if tl.name in self.include_tools]

#         if self.exclude_tools is not None:
#             tool_methods = [tl for tl in tool_methods if tl.name not in self.exclude_tools]

#         # Register filtered tools
#         for tool in tool_methods:
#             # Get the function returned by the tool method
#             if tool is not None:
#                 self.mcp.add_tool(tool)


class BaseMCP:
    """Base class for all Scanpy MCP classes."""

    def __init__(
        self,
        name: str,
        include_tools: list = None,
        exclude_tools: list = None,
        AdataInfo=AdataInfo,
    ):
        """
        Initialize BaseMCP with optional tool filtering.

        Args:
            name (str): Name of the MCP server
            include_tools (list, optional): List of tool names to include. If None, all tools are included.
            exclude_tools (list, optional): List of tool names to exclude. If None, no tools are excluded.
            AdataInfo: The AdataInfo class to use for type annotations.
        """
        self.mcp = FastMCP(name)
        self.include_tools = include_tools
        self.exclude_tools = exclude_tools
        self.AdataInfo = AdataInfo
        self._register_tools()

    def _register_tools(self):
        """Register all tool methods with the FastMCP instance based on include/exclude filters"""
        # Get all methods of the class
        methods = inspect.getmembers(self, predicate=inspect.ismethod)

        # Filter methods that start with _tool_
        tool_methods = [
            tl_method() for name, tl_method in methods if name.startswith("_tool_")
        ]

        # Filter tools based on include/exclude lists
        if self.include_tools is not None:
            tool_methods = [tl for tl in tool_methods if tl.name in self.include_tools]

        if self.exclude_tools is not None:
            tool_methods = [
                tl for tl in tool_methods if tl.name not in self.exclude_tools
            ]

        # Register filtered tools
        for tool in tool_methods:
            # Get the function returned by the tool method
            if tool is not None:
                self.mcp.add_tool(tool)


class BaseMCPManager:
    """Base class for MCP module management."""

    def __init__(
        self,
        name: str,
        instructions: Optional[str] = None,
        include_modules: Optional[List[str]] = None,
        exclude_modules: Optional[List[str]] = None,
        include_tools: Optional[List[str]] = None,
        exclude_tools: Optional[List[str]] = None,
        include_tags: Optional[List[str]] = None,
        exclude_tags: Optional[List[str]] = None,
        backend: Optional[NotebookManager] = None,
    ):
        """
        Initialize BaseMCPManager with optional module filtering.

        Args:
            name (str): Name of the MCP server
            include_modules (List[str], optional): List of module names to include. If None, all modules are included.
            exclude_modules (List[str], optional): List of module names to exclude. If None, no modules are excluded.
            include_tools (List[str], optional): List of tool names to include. If None, all tools are included.
            exclude_tools (List[str], optional): List of tool names to exclude. If None, no tools are excluded.
        """
        self.exec_backend = backend()
        self.mcp = FastMCP(
            name,
            instructions=instructions,
            lifespan=self.exec_lifespan,
            include_tags=include_tags,
            exclude_tags=exclude_tags,
        )
        self.include_modules = include_modules
        self.exclude_modules = exclude_modules
        self.include_tools = include_tools
        self.exclude_tools = exclude_tools
        self.available_modules = {}
        self.init_mcp()
        self.register_mcp()

    def init_mcp(self):
        """Initialize available modules. To be implemented by subclasses."""
        raise NotImplementedError("Subclasses must implement init_mcp")

    def register_mcp(self):
        """Register modules based on include/exclude filters."""
        # Filter modules based on include/exclude lists
        if self.include_modules is not None:
            self.available_modules = {
                k: v
                for k, v in self.available_modules.items()
                if k in self.include_modules
            }

        if self.exclude_modules is not None:
            self.available_modules = {
                k: v
                for k, v in self.available_modules.items()
                if k not in self.exclude_modules
            }

        # Register each module
        for module_name, mcpi in self.available_modules.items():
            if isinstance(mcpi, FastMCP):
                if self.include_tools is not None and module_name in self.include_tools:
                    mcpi = filter_tools(
                        mcpi, include_tools=self.include_tools[module_name]
                    )
                if self.exclude_tools is not None and module_name in self.exclude_tools:
                    mcpi = filter_tools(
                        mcpi, exclude_tools=self.exclude_tools[module_name]
                    )

                asyncio.run(self.mcp.import_server(module_name, mcpi))
            else:
                asyncio.run(self.mcp.import_server(module_name, mcpi().mcp))

    @asynccontextmanager
    async def exec_lifespan(self, server: FastMCP) -> AsyncIterator[Any]:
        yield self.exec_backend



import pytest
from fastmcp import Client
from pathlib import Path
import nest_asyncio

# Apply nest_asyncio at module level
nest_asyncio.apply()


@pytest.mark.asyncio
async def test_complete_workflow(mcp):
    test_dir = Path(__file__).parent / "data/hg19"
    async with Client(mcp) as client:
        # Load the data
        result = await client.call_tool("io_read", {"request": {"filename": test_dir}})
        assert "AnnData" in result.content[0].text

        # Filter cells and genes
        result = await client.call_tool(
            "pp_subset_cells", {"request": {"min_genes": 200}}
        )
        assert "AnnData" in result.content[0].text
        result = await client.call_tool(
            "pp_subset_genes", {"request": {"min_cells": 3}}
        )
        assert "AnnData" in result.content[0].text

        # Add mitochondrial genes and calculate QC metrics
        result = await client.call_tool(
            "ul_mark_var",
            {
                "request": {
                    "var_name": "mt",
                    "pattern_type": "startswith",
                    "patterns": "MT-",
                }
            },
        )
        assert "mt" in result.content[0].text
        result = await client.call_tool(
            "pp_calculate_qc_metrics",
            {"request": {"qc_vars": ["mt"], "percent_top": None, "log1p": False}},
        )
        assert "total_counts" in result.content[0].text

        # Filter based on QC metrics
        result = await client.call_tool(
            "pp_subset_cells", {"request": {"max_genes": 2500, "max_mt_percent": 5}}
        )
        assert "AnnData" in result.content[0].text

        # Normalize and transform
        result = await client.call_tool(
            "pp_normalize_total", {"request": {"target_sum": 1e4}}
        )
        assert "AnnData" in result.content[0].text
        result = await client.call_tool("pp_log1p", {"request": {}})
        assert "log1p" in result.content[0].text

        # Find highly variable genes
        result = await client.call_tool(
            "pp_highly_variable_genes",
            {"request": {"min_mean": 0.0125, "max_mean": 3, "min_disp": 0.5}},
        )
        assert "highly_variable" in result.content[0].text

        # Subset to highly variable genes
        result = await client.call_tool(
            "pp_subset_genes", {"request": {"highly_variable": True}}
        )
        assert "AnnData" in result.content[0].text

        # Regress out effects
        result = await client.call_tool(
            "pp_regress_out", {"request": {"keys": ["total_counts", "pct_counts_mt"]}}
        )
        assert "AnnData" in result.content[0].text

        # Scale the data
        result = await client.call_tool("pp_scale", {"request": {"max_value": 10}})
        assert "AnnData" in result.content[0].text

        # PCA
        result = await client.call_tool(
            "tl_pca", {"request": {"n_comps": 50, "svd_solver": "arpack"}}
        )
        assert "X_pca" in result.content[0].text

        # Compute neighborhood graph
        result = await client.call_tool(
            "pp_neighbors", {"request": {"n_neighbors": 10, "n_pcs": 40}}
        )
        assert "neighbors" in result.content[0].text

        # Leiden clustering
        result = await client.call_tool("tl_leiden", {"request": {}})
        assert "leiden" in result.content[0].text

        # UMAP
        result = await client.call_tool("tl_umap", {"request": {}})
        assert "X_umap" in result.content[0].text

        # Plot UMAP
        result = await client.call_tool(
            "pl_embedding",
            {"request": {"basis": "umap", "color": ["leiden", "NKG7", "PPBP"]}},
        )
        assert "figure" in result.content[0].text



from fastmcp.exceptions import ToolError
from fastmcp.tools.tool import Tool
from scmcp_shared.schema.preset.util import *
from scmcp_shared.schema.preset import AdataInfo
from scmcp_shared.util import forward_request, get_ads, add_op_log
from scmcp_shared.mcp_base import BaseMCP


class ScanpyUtilMCP(BaseMCP):
    def __init__(
        self,
        include_tools: list = None,
        exclude_tools: list = None,
        AdataInfo=AdataInfo,
    ):
        """
        Initialize ScanpyUtilMCP with optional tool filtering.

        Args:
            include_tools (list, optional): List of tool names to include. If None, all tools are included.
            exclude_tools (list, optional): List of tool names to exclude. If None, no tools are excluded.
            AdataInfo: The AdataInfo class to use for type annotations.
        """
        super().__init__("SCMCP-Util-Server", include_tools, exclude_tools, AdataInfo)

    def _tool_query_op_log(self):
        def _query_op_log(
            request: QueryOpLogParam, adinfo: self.AdataInfo = self.AdataInfo()
        ):
            """Query the adata operation log"""
            adata = get_ads().get_adata(adinfo=adinfo)
            op_dic = adata.uns["operation"]["op"]
            opids = adata.uns["operation"]["opid"][-request.n :]
            op_list = []
            for opid in opids:
                op_list.append(op_dic[opid])
            return op_list

        return Tool.from_function(_query_op_log, name="query_op_log", enabled=True)

    def _tool_mark_var(self):
        def _mark_var(request: MarkVarParam, adinfo: self.AdataInfo = self.AdataInfo()):
            """
            Determine if each gene meets specific conditions and store results in adata.var as boolean values.
            For example: mitochondrion genes startswith MT-.
            The tool should be called first when calculate quality control metrics for mitochondrion, ribosomal, harhemoglobin genes, or other qc_vars.
            """
            try:
                result = forward_request("ul_mark_var", request, adinfo)
                if result is not None:
                    return result
                adata = get_ads().get_adata(adinfo=adinfo)
                var_name = request.var_name
                gene_class = request.gene_class
                pattern_type = request.pattern_type
                patterns = request.patterns
                if gene_class is not None:
                    if gene_class == "mitochondrion":
                        adata.var["mt"] = adata.var_names.str.startswith(
                            ("MT-", "Mt", "mt-")
                        )
                        var_name = "mt"
                    elif gene_class == "ribosomal":
                        adata.var["ribo"] = adata.var_names.str.startswith(
                            ("RPS", "RPL", "Rps", "Rpl")
                        )
                        var_name = "ribo"
                    elif gene_class == "hemoglobin":
                        adata.var["hb"] = adata.var_names.str.contains(
                            "^HB[^(P)]", case=False
                        )
                        var_name = "hb"
                elif pattern_type is not None and patterns is not None:
                    if pattern_type == "startswith":
                        adata.var[var_name] = adata.var_names.str.startswith(patterns)
                    elif pattern_type == "endswith":
                        adata.var[var_name] = adata.var_names.str.endswith(patterns)
                    elif pattern_type == "contains":
                        adata.var[var_name] = adata.var_names.str.contains(patterns)
                    else:
                        raise ValueError(
                            f"Did not support pattern_type: {pattern_type}"
                        )
                else:
                    raise ValueError("Please provide validated parameter")

                res = {
                    var_name: adata.var[var_name].value_counts().to_dict(),
                    "msg": f"add '{var_name}' column in adata.var",
                }
                func_kwargs = {
                    "var_name": var_name,
                    "gene_class": gene_class,
                    "pattern_type": pattern_type,
                    "patterns": patterns,
                }
                add_op_log(adata, "mark_var", func_kwargs, adinfo)
                return res
            except ToolError as e:
                raise ToolError(e)
            except Exception as e:
                if hasattr(e, "__context__") and e.__context__:
                    raise ToolError(e.__context__)
                else:
                    raise ToolError(e)

        return Tool.from_function(_mark_var, name="mark_var", enabled=True)

    def _tool_list_var(self):
        def _list_var(
            request: ListVarParam = ListVarParam(),
            adinfo: self.AdataInfo = self.AdataInfo(),
        ):
            """List key columns in adata.var. It should be called for checking when other tools need var key column names as input."""
            try:
                result = forward_request("ul_list_var", request, adinfo)
                if result is not None:
                    return result
                adata = get_ads().get_adata(adinfo=adinfo)
                columns = list(adata.var.columns)
                add_op_log(adata, "list_var", {}, adinfo)
                return columns
            except ToolError as e:
                raise ToolError(e)
            except Exception as e:
                if hasattr(e, "__context__") and e.__context__:
                    raise ToolError(e.__context__)
                else:
                    raise ToolError(e)

        return Tool.from_function(_list_var, name="list_var", enabled=True)

    def _tool_list_obs(self):
        def _list_obs(request: ListObsParam, adinfo: self.AdataInfo = self.AdataInfo()):
            """List key columns in adata.obs. It should be called before other tools need obs key column names input."""
            try:
                result = forward_request("ul_list_obs", request, adinfo)
                if result is not None:
                    return result
                adata = get_ads().get_adata(adinfo=adinfo)
                columns = list(adata.obs.columns)
                add_op_log(adata, "list_obs", {}, adinfo)
                return columns
            except ToolError as e:
                raise ToolError(e)
            except Exception as e:
                if hasattr(e, "__context__") and e.__context__:
                    raise ToolError(e.__context__)
                else:
                    raise ToolError(e)

        return Tool.from_function(_list_obs, name="list_obs", enabled=True)

    def _tool_check_var(self):
        def _check_var(
            request: VarNamesParam, adinfo: self.AdataInfo = self.AdataInfo()
        ):
            """Check if genes/variables exist in adata.var_names. This tool should be called before gene expression visualizations or color by genes."""
            try:
                result = forward_request("ul_check_var", request, adinfo)
                if result is not None:
                    return result
                adata = get_ads().get_adata(adinfo=adinfo)
                var_names = request.var_names
                if adata.raw is not None:
                    all_var_names = adata.raw.to_adata().var_names
                else:
                    all_var_names = adata.var_names
                result = {v: v in all_var_names for v in var_names}
                add_op_log(adata, "check_var", {"var_names": var_names}, adinfo)
                return result
            except ToolError as e:
                raise ToolError(e)
            except Exception as e:
                if hasattr(e, "__context__") and e.__context__:
                    raise ToolError(e.__context__)
                else:
                    raise ToolError(e)

        return Tool.from_function(_check_var, name="check_var", enabled=True)

    def _tool_merge_adata(self):
        def _merge_adata(
            request: ConcatBaseParam, adinfo: self.AdataInfo = self.AdataInfo()
        ):
            """Merge multiple adata objects."""
            try:
                result = forward_request("ul_merge_adata", request, adinfo)
                if result is not None:
                    return result
                ads = get_ads()
                adata = ads.get_adata(adinfo=adinfo)
                kwargs = {
                    k: v for k, v in request.model_dump().items() if v is not None
                }
                merged_adata = adata.concat(ads.adata_dic, **kwargs)
                ads.adata_dic[dtype] = {}
                ads.active_id = "merged_adata"
                add_op_log(merged_adata, ad.concat, kwargs, adinfo)
                ads.adata_dic[ads.active_id] = merged_adata
                return {
                    "status": "success",
                    "message": "Successfully merged all AnnData objects",
                }
            except ToolError as e:
                raise ToolError(e)
            except Exception as e:
                if hasattr(e, "__context__") and e.__context__:
                    raise ToolError(e.__context__)
                else:
                    raise ToolError(e)

        return Tool.from_function(_merge_adata, name="merge_adata", enabled=True)

    def _tool_set_dpt_iroot(self):
        def _set_dpt_iroot(
            request: DPTIROOTParam, adinfo: self.AdataInfo = self.AdataInfo()
        ):
            """Set the iroot cell"""
            try:
                result = forward_request("ul_set_dpt_iroot", request, adinfo)
                if result is not None:
                    return result
                adata = get_ads().get_adata(adinfo=adinfo)
                diffmap_key = request.diffmap_key
                dimension = request.dimension
                direction = request.direction
                if diffmap_key not in adata.obsm:
                    raise ValueError(
                        f"Diffusion map key '{diffmap_key}' not found in adata.obsm"
                    )
                if direction == "min":
                    adata.uns["iroot"] = adata.obsm[diffmap_key][:, dimension].argmin()
                else:
                    adata.uns["iroot"] = adata.obsm[diffmap_key][:, dimension].argmax()

                func_kwargs = {
                    "diffmap_key": diffmap_key,
                    "dimension": dimension,
                    "direction": direction,
                }
                add_op_log(adata, "set_dpt_iroot", func_kwargs, adinfo)

                return {
                    "status": "success",
                    "message": f"Successfully set root cell for DPT using {direction} of dimension {dimension}",
                }
            except ToolError as e:
                raise ToolError(e)
            except Exception as e:
                if hasattr(e, "__context__") and e.__context__:
                    raise ToolError(e.__context__)
                else:
                    raise ToolError(e)

        return Tool.from_function(_set_dpt_iroot, name="set_dpt_iroot", enabled=True)

    def _tool_add_layer(self):
        def _add_layer(
            request: AddLayerParam, adinfo: self.AdataInfo = self.AdataInfo()
        ):
            """Add a layer to the AnnData object."""
            try:
                result = forward_request("ul_add_layer", request, adinfo)
                if result is not None:
                    return result
                adata = get_ads().get_adata(adinfo=adinfo)
                layer_name = request.layer_name

                # Check if layer already exists
                if layer_name in adata.layers:
                    raise ValueError(
                        f"Layer '{layer_name}' already exists in adata.layers"
                    )
                # Add the data as a new layer
                adata.layers[layer_name] = adata.X.copy()

                func_kwargs = {"layer_name": layer_name}
                add_op_log(adata, "add_layer", func_kwargs, adinfo)

                return {
                    "status": "success",
                    "message": f"Successfully added layer '{layer_name}' to adata.layers",
                }
            except ToolError as e:
                raise ToolError(e)
            except Exception as e:
                if hasattr(e, "__context__") and e.__context__:
                    raise ToolError(e.__context__)
                else:
                    raise ToolError(e)

        return Tool.from_function(_add_layer, name="add_layer", enabled=True)

    def _tool_check_samples(self):
        def _check_samples(request: None, adinfo: self.AdataInfo = self.AdataInfo()):
            """check the stored samples"""
            try:
                ads = get_ads()
                return {
                    "sampleid": [
                        list(ads.adata_dic[dk].keys()) for dk in ads.adata_dic.keys()
                    ]
                }
            except ToolError as e:
                raise ToolError(e)
            except Exception as e:
                if hasattr(e, "__context__") and e.__context__:
                    raise ToolError(e.__context__)
                else:
                    raise ToolError(e)

        return Tool.from_function(_check_samples, name="check_samples", enabled=True)



from pydantic import BaseModel, Field
from typing import List


class Tool(BaseModel):
    name: str = Field(description="The name of the tool")
    description: str = Field(description="The description of the tool")


class ToolList(BaseModel):
    tools: List[Tool]


from agno.document.chunking.agentic import AgenticChunking
from agno.embedder.openai import OpenAIEmbedder
from agno.models.deepseek import DeepSeek
from agno.vectordb.lancedb import LanceDb
from agno.knowledge.agent import AgentKnowledge
import importlib.resources
import os
import requests
import zipfile
import tempfile
import shutil
from pathlib import Path
import logging

logger = logging.getLogger(__name__)

embedder_id = os.getenv("EMBEDDER_MODEL")
embedder_api_key = os.getenv("EMBEDDER_API_KEY")
embedder_base_url = os.getenv("EMBEDDER_BASE_URL")
model_id = os.getenv("MODEL")
model_api_key = os.getenv("API_KEY")
model_base_url = os.getenv("BASE_URL")

# 配置信息
config = {
    "local_dir": "vector_db",
    "huggingface_url": "https://huggingface.co/datasets/huangshing/scmcp_vector_db/resolve/main/vector_db.zip",
}


def download_vector_db(source="huggingface"):
    """
    下载向量数据库文件

    Args:
        source: 下载源 ("huggingface" 或 "github")
    """

    # 获取本地存储路径
    package_path = importlib.resources.path("scmcp_shared", "")
    local_dir = Path(package_path) / config["local_dir"]
    local_dir.mkdir(exist_ok=True)

    # 检查是否已存在
    if (local_dir / "scmcp.lance").exists():
        logger.info("Vector database already exists locally")
        return str(local_dir)

    logger.info(f"Downloading vector database from {source}...")

    # 创建临时目录用于下载和解压
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_path = Path(temp_dir)
        zip_path = temp_path / "vector_db.zip"

        try:
            # 下载文件
            if source == "huggingface":
                url = config["huggingface_url"]
            else:
                raise ValueError(f"Unsupported source: {source}")

            logger.info(f"Downloading from: {url}")
            response = requests.get(url, stream=True)
            response.raise_for_status()

            with open(zip_path, "wb") as f:
                for chunk in response.iter_content(chunk_size=8192):
                    f.write(chunk)

            # 解压文件
            logger.info("Extracting downloaded archive...")
            _extract_archive(zip_path, local_dir)

            logger.info(f"Vector database downloaded and extracted to: {local_dir}")
            return str(local_dir)

        except requests.RequestException as e:
            raise RuntimeError(f"Failed to download vector database: {e}")
        except Exception as e:
            raise RuntimeError(f"Failed to process vector database: {e}")


def _extract_archive(archive_path, extract_dir):
    """解压归档文件"""
    with zipfile.ZipFile(archive_path, "r") as zip_ref:
        zip_ref.extractall(extract_dir)

    # 如果解压后只有一个子目录，移动内容到目标目录
    extracted_items = list(Path(extract_dir).iterdir())
    if len(extracted_items) == 1 and extracted_items[0].is_dir():
        subdir = extracted_items[0]
        for item in subdir.iterdir():
            shutil.move(str(item), str(extract_dir / item.name))
        subdir.rmdir()


def load_kb(software=None, auto_download=True, download_source="huggingface"):
    """
    加载知识库

    Args:
        software: 软件名称
        auto_download: 是否自动下载向量数据库
        download_source: 下载源 ("huggingface" 或 "github")
    """
    # 获取向量数据库路径
    try:
        vector_db_path = importlib.resources.path("scmcp_shared", "vector_db")
    except FileNotFoundError:
        if auto_download:
            logger.info("Vector database not found in package, attempting download...")
            vector_db_path = download_vector_db(download_source)
        else:
            raise FileNotFoundError(
                "Vector database not found. Set auto_download=True to download automatically, "
                "or manually place the vector database in the scmcp_shared package."
            )

    vector_db = LanceDb(
        table_name=software,
        uri=vector_db_path,
        embedder=OpenAIEmbedder(
            id=embedder_id,
            base_url=embedder_base_url,
            api_key=embedder_api_key,
        ),
    )
    model = DeepSeek(
        id=model_id,
        base_url=model_base_url,
        api_key=model_api_key,
    )
    knowledge_base = AgentKnowledge(
        chunking_strategy=AgenticChunking(model=model),
        vector_db=vector_db,
    )

    return knowledge_base



from __future__ import annotations

from pydantic import Field, BaseModel, ConfigDict


class AdataInfo(BaseModel):
    """Input schema for the adata tool."""

    sampleid: str | None = Field(default=None, description="adata sampleid")
    adtype: str = Field(
        default="exp",
        description="The input adata.X data type for preprocess/analysis/plotting",
    )

    model_config = ConfigDict(extra="ignore")



import inspect
import os
from pathlib import Path
from fastmcp.server.dependencies import get_context
from fastmcp.exceptions import ToolError
import asyncio
import nest_asyncio
from pydantic import BaseModel


def get_env(key):
    return os.environ.get(f"SCMCP_{key.upper()}")


def filter_args(request, func, **extra_kwargs):
    kwargs = request.model_dump()
    args = request.model_fields_set
    parameters = inspect.signature(func).parameters
    extra_kwargs = {k: extra_kwargs[k] for k in extra_kwargs if k in parameters}
    func_kwargs = {k: kwargs.get(k) for k in args if k in parameters}
    func_kwargs.update(extra_kwargs)
    return func_kwargs


def add_op_log(adata, func, kwargs, adinfo):
    import hashlib
    import json

    if "operation" not in adata.uns:
        adata.uns["operation"] = {}
        adata.uns["operation"]["op"] = {}
        adata.uns["operation"]["opid"] = []
    # Handle different function types to get the function name
    if hasattr(func, "func") and hasattr(func.func, "__name__"):
        # For partial functions, use the original function name
        func_name = func.func.__name__
    elif hasattr(func, "__name__"):
        func_name = func.__name__
    elif hasattr(func, "__class__"):
        func_name = func.__class__.__name__
    else:
        func_name = str(func)
    new_kwargs = {**adinfo.model_dump()}
    for k, v in kwargs.items():
        if isinstance(v, tuple):
            new_kwargs[k] = list(v)
        else:
            new_kwargs[k] = v
    try:
        kwargs_str = json.dumps(new_kwargs, sort_keys=True)
    except Exception as e:
        print(e)
        kwargs_str = f"{e}" + str(new_kwargs)
    hash_input = f"{func_name}:{kwargs_str}"
    hash_key = hashlib.md5(hash_input.encode()).hexdigest()
    adata.uns["operation"]["op"][hash_key] = {func_name: new_kwargs}
    adata.uns["operation"]["opid"] = list(adata.uns["operation"]["opid"])
    adata.uns["operation"]["opid"].append(hash_key)
    from .logging_config import setup_logger

    logger = setup_logger(log_file=get_env("LOG_FILE"))
    logger.info(f"{func}: {new_kwargs}")


def save_fig_path(axes, file):
    from matplotlib.axes import Axes

    try:
        file_path = Path(file)
        file_path.parent.mkdir(parents=True, exist_ok=True)
        if isinstance(axes, list):
            if isinstance(axes[0], Axes):
                axes[0].figure.savefig(file_path)
        elif isinstance(axes, dict):
            ax = list(axes.values())[0]
            if isinstance(ax, Axes):
                ax.figure.savefig(file_path)
        elif isinstance(axes, Axes):
            axes.figure.savefig(file_path)
        elif hasattr(axes, "savefig"):  # if Figure
            axes.savefig(file_path)
        elif hasattr(axes, "save"):  # for plotnine.ggplot.ggplot
            axes.save(file_path)
        else:
            raise ValueError(
                f"axes must be a Axes or plotnine object, but got {type(axes)}"
            )
        return file_path
    except Exception as e:
        raise e


def savefig(axes, func=None, **kwargs):
    if hasattr(func, "func") and hasattr(func.func, "__name__"):
        # For partial functions, use the original function name
        func_name = func.func.__name__
    elif hasattr(func, "__name__"):
        func_name = func.__name__
    elif hasattr(func, "__class__"):
        func_name = func.__class__.__name__
    else:
        func_name = str(func)

    fig_dir = Path(os.getcwd()) / "figures"
    kwargs.pop("save", None)
    kwargs.pop("show", None)
    args = []
    for k, v in kwargs.items():
        if isinstance(v, (tuple, list, set)):
            v = v[:3]  ## show first 3 elements
            args.append(f"{k}-{'-'.join([str(i) for i in v])}")
        elif isinstance(v, dict):
            v = list(v.keys())[:3]  ## show first 3 elements
            args.append(f"{k}-{'-'.join([str(i) for i in v])}")
        else:
            args.append(f"{k}-{v}")
    args_str = "_".join(args).replace(" ", "")
    fig_path = fig_dir / f"{func_name}_{args_str}.png"
    try:
        save_fig_path(axes, fig_path)
    except PermissionError:
        raise PermissionError("You don't have permission to save figure")
    except Exception as e:
        raise e
    transport = get_env("TRANSPORT")
    if transport == "stdio":
        return fig_path
    else:
        host = get_env("HOST")
        port = get_env("PORT")
        fig_path = f"http://{host}:{port}/figures/{Path(fig_path).name}"
        return fig_path


async def get_figure(request):
    from starlette.responses import FileResponse, Response

    figure_name = request.path_params["figure_name"]
    figure_path = f"./figures/{figure_name}"

    if not os.path.isfile(figure_path):
        return Response(
            content={"error": "figure not found"}, media_type="application/json"
        )

    return FileResponse(figure_path)


def add_figure_route(server):
    from starlette.routing import Route

    server._additional_http_routes = [
        Route("/figures/{figure_name}", endpoint=get_figure)
    ]


async def async_forward_request(func, request, adinfo, **kwargs):
    from fastmcp import Client

    forward_url = get_env("FORWARD")
    request_kwargs = request.model_dump()
    request_args = request.model_fields_set
    func_kwargs = {
        "request": {k: request_kwargs.get(k) for k in request_args},
        "adinfo": adinfo.model_dump(),
    }
    if not forward_url:
        return None

    client = Client(forward_url)
    async with client:
        tools = await client.list_tools()
        func = [t.name for t in tools if t.name.endswith(func)][0]
        try:
            result = await client.call_tool(func, func_kwargs)
            return result
        except ToolError as e:
            raise ToolError(e)
        except Exception as e:
            if hasattr(e, "__context__") and e.__context__:
                raise Exception(f"{str(e.__context__)}")
            else:
                raise e


def forward_request(func, request, adinfo, **kwargs):
    """Synchronous wrapper for forward_request"""
    try:
        # Apply nest_asyncio to allow nested event loops
        nest_asyncio.apply()
        loop = asyncio.get_event_loop()
        if loop.is_running():
            # If we're in a running event loop, use create_task
            async def _run():
                return await async_forward_request(func, request, adinfo, **kwargs)

            return loop.run_until_complete(_run())
        else:
            # If no event loop is running, use asyncio.run()
            return asyncio.run(async_forward_request(func, request, adinfo, **kwargs))
    except Exception as e:
        if hasattr(e, "__context__") and e.__context__:
            raise Exception(f"{str(e.__context__)}")
        else:
            raise e


def obsm2adata(adata, obsm_key):
    from anndata import AnnData

    if obsm_key not in adata.obsm_keys():
        raise ValueError(f"key {obsm_key} not found in adata.obsm")
    else:
        return AnnData(
            adata.obsm[obsm_key], obs=adata.obs, obsm=adata.obsm, uns=adata.uns
        )


def get_ads():
    ctx = get_context()
    ads = ctx.request_context.lifespan_context
    return ads


def generate_msg(adinfo, adata, ads):
    return {
        "sampleid": adinfo.sampleid or ads.active_id,
        "adtype": adinfo.adtype,
        "adata": adata,
    }


def sc_like_plot(plot_func, adata, request, adinfo, **kwargs):
    from matplotlib import pyplot as plt

    func_kwargs = filter_args(request, plot_func, show=False, save=False)
    axes = plot_func(adata, **func_kwargs)
    if axes is None:
        axes = plt.gca()
    fig_path = savefig(axes, plot_func, **func_kwargs)
    add_op_log(adata, plot_func, func_kwargs, adinfo)
    return fig_path


def filter_tools(mcp, include_tools=None, exclude_tools=None):
    import asyncio
    import copy

    mcp = copy.deepcopy(mcp)

    async def _filter_tools(mcp, include_tools=None, exclude_tools=None):
        tools = await mcp.get_tools()
        for tool in tools:
            if exclude_tools and tool in exclude_tools:
                mcp.remove_tool(tool)
            if include_tools and tool not in include_tools:
                mcp.remove_tool(tool)
        return mcp

    return asyncio.run(_filter_tools(mcp, include_tools, exclude_tools))


def set_env(log_file, forward, transport, host, port):
    if log_file is not None:
        os.environ["SCMCP_LOG_FILE"] = log_file
    if forward is not None:
        os.environ["SCMCP_FORWARD"] = forward
    os.environ["SCMCP_TRANSPORT"] = transport
    os.environ["SCMCP_HOST"] = host
    os.environ["SCMCP_PORT"] = str(port)


def setup_mcp(mcp, sub_mcp_dic, modules=None):
    import asyncio

    if modules is None or modules == "all":
        modules = sub_mcp_dic.keys()
    for module in modules:
        asyncio.run(mcp.import_server(module, sub_mcp_dic[module]))
    return mcp


def _update_args(mcp, func, args_dic: dict):
    for args, property_dic in args_dic.items():
        for pk, v in property_dic.items():
            mcp._tool_manager._tools[func].parameters["properties"][
                "request"
            ].setdefault(pk, {})
            mcp._tool_manager._tools[func].parameters["properties"]["request"][pk][
                args
            ] = v


def update_mcp_args(mcp, tool_args: dict):
    # tools = mcp._tool_manager._tools.keys()
    for tool in tool_args:
        _update_args(mcp, tool, tool_args[tool])


def check_adata(adata, adinfo, ads):
    sampleid = adinfo.sampleid or ads.active_id
    if sampleid != adata.uns["scmcp_sampleid"]:
        raise ValueError(
            f"sampleid mismatch: {sampleid} != {adata.uns['scmcp_sampleid']}"
        )


def get_nbm():
    ctx = get_context()
    nbm = ctx.request_context.lifespan_context
    return nbm


def parse_args(
    kwargs: BaseModel | dict,
    positional_args: list[str] | str | None = None,
    func_args: list[str] | str | None = None,
) -> str:
    if isinstance(kwargs, BaseModel):
        kwargs = kwargs.model_dump()
    elif isinstance(kwargs, dict):
        kwargs = kwargs
    else:
        raise ValueError(f"Invalid type: {type(kwargs)}")

    if func_args is not None:
        if isinstance(func_args, str):
            func_args = [func_args]
    else:
        func_args = []
    kwargs_str_ls = []
    if positional_args is not None:
        if isinstance(positional_args, str):
            kwargs_str_ls.append(kwargs.pop(positional_args))
        elif isinstance(positional_args, list):
            for arg in positional_args:
                kwargs_str_ls.append(kwargs.pop(arg))

    extra_kwargs = kwargs.pop("kwargs", {})
    for k, v in kwargs.items():
        if k in func_args:
            kwargs_str_ls.append(f"{k}={v}")
            continue
        if isinstance(v, (list, tuple, dict, int, float, bool)):
            kwargs_str_ls.append(f"{k}={v}")
        elif isinstance(v, str):
            kwargs_str_ls.append(f"{k}='{v}'")

    if extra_kwargs:
        kwargs_str_ls.append(f"**{extra_kwargs}")

    kwargs_str = ", ".join(kwargs_str_ls)
    return kwargs_str



import scanpy as sc
from fastmcp.tools.tool import Tool
from fastmcp.exceptions import ToolError
from scmcp_shared.schema.preset.pl import *
from scmcp_shared.schema.preset import AdataInfo
from scmcp_shared.util import forward_request, sc_like_plot, get_ads
from scmcp_shared.mcp_base import BaseMCP


class ScanpyPlottingMCP(BaseMCP):
    def __init__(
        self,
        include_tools: list = None,
        exclude_tools: list = None,
        AdataInfo: AdataInfo = AdataInfo,
    ):
        """
        Initialize ScanpyPreprocessingMCP with optional tool filtering.

        Args:
            include_tools (list, optional): List of tool names to include. If None, all tools are included.
            exclude_tools (list, optional): List of tool names to exclude. If None, no tools are excluded.
            AdataInfo: The AdataInfo class to use for type annotations.
        """
        super().__init__("ScanpyMCP-PL-Server", include_tools, exclude_tools, AdataInfo)

    def _tool_pca(self):
        def _pca(request: PCAParam, adinfo: self.AdataInfo = self.AdataInfo()):
            """Scatter plot in PCA coordinates. default figure for PCA plot"""
            try:
                if (res := forward_request("pl_pca", request, adinfo)) is not None:
                    return res
                adata = get_ads().get_adata(adinfo=adinfo)
                fig_path = sc_like_plot(sc.pl.pca, adata, request, adinfo)
                return {"figpath": fig_path}
            except ToolError as e:
                raise ToolError(e)
            except Exception as e:
                if hasattr(e, "__context__") and e.__context__:
                    raise ToolError(e.__context__)
                else:
                    raise ToolError(e)

        return Tool.from_function(_pca, name="pca", enabled=True)

    def _tool_diffmap(self):
        def _diffmap(
            request: DiffusionMapParam, adinfo: self.AdataInfo = self.AdataInfo()
        ):
            """Plot diffusion map embedding of cells."""
            try:
                if (res := forward_request("pl_diffmap", request, adinfo)) is not None:
                    return res
                adata = get_ads().get_adata(adinfo=adinfo)
                fig_path = sc_like_plot(sc.pl.diffmap, adata, request, adinfo)
                return {"figpath": fig_path}
            except ToolError as e:
                raise ToolError(e)
            except Exception as e:
                if hasattr(e, "__context__") and e.__context__:
                    raise ToolError(e.__context__)
                else:
                    raise ToolError(e)

        return Tool.from_function(_diffmap, name="diffmap", enabled=True)

    def _tool_violin(self):
        def _violin(request: ViolinParam, adinfo: self.AdataInfo = self.AdataInfo()):
            """Plot violin plot of one or more variables."""
            try:
                if (res := forward_request("pl_violin", request, adinfo)) is not None:
                    return res
                adata = get_ads().get_adata(adinfo=adinfo)
                fig_path = sc_like_plot(sc.pl.violin, adata, request, adinfo)
                return {"figpath": fig_path}
            except KeyError as e:
                raise ToolError(
                    f"doest found {e} in current sampleid with adtype {adinfo.adtype}"
                )
            except ToolError as e:
                raise ToolError(e)
            except Exception as e:
                if hasattr(e, "__context__") and e.__context__:
                    raise ToolError(e.__context__)
                else:
                    raise ToolError(e)

        return Tool.from_function(_violin, name="violin", enabled=True)

    def _tool_stacked_violin(self):
        def _stacked_violin(
            request: StackedViolinParam, adinfo: self.AdataInfo = self.AdataInfo()
        ):
            """Plot stacked violin plots. Makes a compact image composed of individual violin plots stacked on top of each other."""
            try:
                if (
                    res := forward_request("pl_stacked_violin", request, adinfo)
                ) is not None:
                    return res
                adata = get_ads().get_adata(adinfo=adinfo)
                fig_path = sc_like_plot(sc.pl.stacked_violin, adata, request, adinfo)
                return {"figpath": fig_path}
            except ToolError as e:
                raise ToolError(e)
            except Exception as e:
                if hasattr(e, "__context__") and e.__context__:
                    raise ToolError(e.__context__)
                else:
                    raise ToolError(e)

        return Tool.from_function(_stacked_violin, name="stacked_violin", enabled=True)

    def _tool_heatmap(self):
        async def _heatmap(
            request: HeatmapParam, adinfo: self.AdataInfo = self.AdataInfo()
        ):
            """Heatmap of the expression values of genes."""
            try:
                if (res := forward_request("pl_heatmap", request, adinfo)) is not None:
                    return res
                adata = get_ads().get_adata(adinfo=adinfo)
                fig_path = sc_like_plot(sc.pl.heatmap, adata, request, adinfo)
                return {"figpath": fig_path}
            except ToolError as e:
                raise ToolError(e)
            except Exception as e:
                if hasattr(e, "__context__") and e.__context__:
                    raise ToolError(e.__context__)
                else:
                    raise ToolError(e)

        return Tool.from_function(_heatmap, name="heatmap", enabled=True)

    def _tool_dotplot(self):
        def _dotplot(request: DotplotParam, adinfo: self.AdataInfo = self.AdataInfo()):
            """Plot dot plot of expression values per gene for each group."""
            try:
                if (res := forward_request("pl_dotplot", request, adinfo)) is not None:
                    return res
                adata = get_ads().get_adata(adinfo=adinfo)
                fig_path = sc_like_plot(sc.pl.dotplot, adata, request, adinfo)
                return {"figpath": fig_path}
            except ToolError as e:
                raise ToolError(e)
            except Exception as e:
                if hasattr(e, "__context__") and e.__context__:
                    raise ToolError(e.__context__)
                else:
                    raise ToolError(e)

        return Tool.from_function(_dotplot, name="dotplot", enabled=True)

    def _tool_matrixplot(self):
        def _matrixplot(
            request: MatrixplotParam, adinfo: self.AdataInfo = self.AdataInfo()
        ):
            """matrixplot, Create a heatmap of the mean expression values per group of each var_names."""
            try:
                if (
                    res := forward_request("pl_matrixplot", request, adinfo)
                ) is not None:
                    return res
                adata = get_ads().get_adata(adinfo=adinfo)
                fig_path = sc_like_plot(sc.pl.matrixplot, adata, request, adinfo)
                return {"figpath": fig_path}
            except ToolError as e:
                raise ToolError(e)
            except Exception as e:
                if hasattr(e, "__context__") and e.__context__:
                    raise ToolError(e.__context__)
                else:
                    raise ToolError(e)

        return Tool.from_function(_matrixplot, name="matrixplot", enabled=True)

    def _tool_tracksplot(self):
        def _tracksplot(
            request: TracksplotParam, adinfo: self.AdataInfo = self.AdataInfo()
        ):
            """tracksplot, compact plot of expression of a list of genes."""
            try:
                if (
                    res := forward_request("pl_tracksplot", request, adinfo)
                ) is not None:
                    return res
                adata = get_ads().get_adata(adinfo=adinfo)
                fig_path = sc_like_plot(sc.pl.tracksplot, adata, request, adinfo)
                return {"figpath": fig_path}
            except ToolError as e:
                raise ToolError(e)
            except Exception as e:
                if hasattr(e, "__context__") and e.__context__:
                    raise ToolError(e.__context__)
                else:
                    raise ToolError(e)

        return Tool.from_function(_tracksplot, name="tracksplot")

    def _tool_scatter(self):
        def _scatter(
            request: EnhancedScatterParam, adinfo: self.AdataInfo = self.AdataInfo()
        ):
            """Plot a scatter plot of two variables, Scatter plot along observations or variables axes."""
            try:
                if (res := forward_request("pl_scatter", request, adinfo)) is not None:
                    return res
                adata = get_ads().get_adata(adinfo=adinfo)
                fig_path = sc_like_plot(sc.pl.scatter, adata, request, adinfo)
                return {"figpath": fig_path}
            except ToolError as e:
                raise ToolError(e)
            except Exception as e:
                if hasattr(e, "__context__") and e.__context__:
                    raise ToolError(e.__context__)
                else:
                    raise ToolError(e)

        return Tool.from_function(_scatter, name="scatter", enabled=True)

    def _tool_embedding(self):
        def _embedding(
            request: EmbeddingParam, adinfo: self.AdataInfo = self.AdataInfo()
        ):
            """Scatter plot for user specified embedding basis (e.g. umap, tsne, etc)."""
            try:
                if (
                    res := forward_request("pl_embedding", request, adinfo)
                ) is not None:
                    return res
                adata = get_ads().get_adata(adinfo=adinfo)
                fig_path = sc_like_plot(sc.pl.embedding, adata, request, adinfo)
                return {"figpath": fig_path}
            except KeyError as e:
                raise ToolError(
                    f"doest found {e} in current sampleid with adtype {adinfo.adtype}"
                )
            except Exception as e:
                if hasattr(e, "__context__") and e.__context__:
                    raise ToolError(e.__context__)
                else:
                    raise ToolError(e)

        return Tool.from_function(_embedding, name="embedding", enabled=True)

    def _tool_embedding_density(self):
        def _embedding_density(
            request: EmbeddingDensityParam, adinfo: self.AdataInfo = self.AdataInfo()
        ):
            """Plot the density of cells in an embedding."""
            try:
                if (
                    res := forward_request("pl_embedding_density", request, adinfo)
                ) is not None:
                    return res
                adata = get_ads().get_adata(adinfo=adinfo)
                fig_path = sc_like_plot(sc.pl.embedding_density, adata, request, adinfo)
                return {"figpath": fig_path}
            except ToolError as e:
                raise ToolError(e)
            except Exception as e:
                if hasattr(e, "__context__") and e.__context__:
                    raise ToolError(e.__context__)
                else:
                    raise ToolError(e)

        return Tool.from_function(
            _embedding_density, name="embedding_density", enabled=True
        )

    def _tool_rank_genes_groups(self):
        def _rank_genes_groups(
            request: RankGenesGroupsParam, adinfo: self.AdataInfo = self.AdataInfo()
        ):
            """Plot ranking of genes based on differential expression."""
            try:
                if (
                    res := forward_request("pl_rank_genes_groups", request, adinfo)
                ) is not None:
                    return res
                adata = get_ads().get_adata(adinfo=adinfo)
                fig_path = sc_like_plot(sc.pl.rank_genes_groups, adata, request, adinfo)
                return {"figpath": fig_path}
            except ToolError as e:
                raise ToolError(e)
            except Exception as e:
                if hasattr(e, "__context__") and e.__context__:
                    raise ToolError(e.__context__)
                else:
                    raise ToolError(e)

        return Tool.from_function(
            _rank_genes_groups, name="rank_genes_groups", enabled=True
        )

    def _tool_rank_genes_groups_dotplot(self):
        def _rank_genes_groups_dotplot(
            request: RankGenesGroupsDotplotParam,
            adinfo: self.AdataInfo = self.AdataInfo(),
        ):
            """Plot ranking of genes(DEGs) using dotplot visualization. Defualt plot DEGs for rank_genes_groups tool"""
            try:
                if (
                    res := forward_request(
                        "pl_rank_genes_groups_dotplot", request, adinfo
                    )
                ) is not None:
                    return res
                adata = get_ads().get_adata(adinfo=adinfo)
                fig_path = sc_like_plot(
                    sc.pl.rank_genes_groups_dotplot, adata, request, adinfo
                )
                return {"figpath": fig_path}
            except ToolError as e:
                raise ToolError(e)
            except Exception as e:
                if hasattr(e, "__context__") and e.__context__:
                    raise ToolError(e.__context__)
                else:
                    raise ToolError(e)

        return Tool.from_function(
            _rank_genes_groups_dotplot, name="rank_genes_groups_dotplot", enabled=True
        )

    def _tool_clustermap(self):
        def _clustermap(
            request: ClusterMapParam, adinfo: self.AdataInfo = self.AdataInfo()
        ):
            """Plot hierarchical clustering of cells and genes."""
            try:
                if (
                    res := forward_request("pl_clustermap", request, adinfo)
                ) is not None:
                    return res
                adata = get_ads().get_adata(adinfo=adinfo)
                fig_path = sc_like_plot(sc.pl.clustermap, adata, request, adinfo)
                return {"figpath": fig_path}
            except ToolError as e:
                raise ToolError(e)
            except Exception as e:
                if hasattr(e, "__context__") and e.__context__:
                    raise ToolError(e.__context__)
                else:
                    raise ToolError(e)

        return Tool.from_function(_clustermap, name="clustermap", enabled=True)

    def _tool_highly_variable_genes(self):
        def _highly_variable_genes(
            request: HighlyVariableGenesParam, adinfo: self.AdataInfo = self.AdataInfo()
        ):
            """plot highly variable genes; Plot dispersions or normalized variance versus means for genes."""
            try:
                if (
                    res := forward_request("pl_highly_variable_genes", request, adinfo)
                ) is not None:
                    return res
                adata = get_ads().get_adata(adinfo=adinfo)
                fig_path = sc_like_plot(
                    sc.pl.highly_variable_genes, adata, request, adinfo
                )
                return {"figpath": fig_path}
            except ToolError as e:
                raise ToolError(e)
            except Exception as e:
                if hasattr(e, "__context__") and e.__context__:
                    raise ToolError(e.__context__)
                else:
                    raise ToolError(e)

        return Tool.from_function(
            _highly_variable_genes, name="highly_variable_genes", enabled=True
        )

    def _tool_pca_variance_ratio(self):
        def _pca_variance_ratio(
            request: PCAVarianceRatioParam, adinfo: self.AdataInfo = self.AdataInfo()
        ):
            """Plot the PCA variance ratio to visualize explained variance."""
            try:
                if (
                    res := forward_request("pl_pca_variance_ratio", request, adinfo)
                ) is not None:
                    return res
                adata = get_ads().get_adata(adinfo=adinfo)
                fig_path = sc_like_plot(
                    sc.pl.pca_variance_ratio, adata, request, adinfo
                )
                return {"figpath": fig_path}
            except ToolError as e:
                raise ToolError(e)
            except Exception as e:
                if hasattr(e, "__context__") and e.__context__:
                    raise ToolError(e.__context__)
                else:
                    raise ToolError(e)

        return Tool.from_function(
            _pca_variance_ratio, name="pca_variance_ratio", enabled=True
        )



import pytest
from fastmcp import Client
from pathlib import Path
import nest_asyncio

# Apply nest_asyncio at module level
nest_asyncio.apply()


@pytest.mark.asyncio
async def test_read_and_write(mcp):
    # Pass the server directly to the Client constructor
    test_dir = Path(__file__).parent / "data/hg19"
    outfile = Path(__file__).parent / "data/test.h5ad"
    async with Client(mcp) as client:
        # tools = await client.list_tools()
        # client._tools = tools
        result = await client.call_tool("io_read", {"request": {"filename": test_dir}})
        print(result)
        assert "AnnData" in result.content[0].text

        result = await client.call_tool("io_write", {"request": {"filename": outfile}})
        assert outfile.exists()
        outfile.unlink()



from pydantic import Field, field_validator, BaseModel

from typing import Optional, Union, List, Dict, Any
from typing import Literal
import numpy as np


class FilterCells(BaseModel):
    """Input schema for the filter_cells preprocessing tool."""

    min_counts: Optional[int] = Field(
        default=None,
        description="Minimum number of counts required for a cell to pass filtering.",
    )

    min_genes: Optional[int] = Field(
        default=None,
        description="Minimum number of genes expressed required for a cell to pass filtering.",
    )

    max_counts: Optional[int] = Field(
        default=None,
        description="Maximum number of counts required for a cell to pass filtering.",
    )

    max_genes: Optional[int] = Field(
        default=None,
        description="Maximum number of genes expressed required for a cell to pass filtering.",
    )

    @field_validator("min_counts", "min_genes", "max_counts", "max_genes")
    def validate_positive_integers(cls, v: Optional[int]) -> Optional[int]:
        """验证整数参数为正数"""
        if v is not None and v <= 0:
            raise ValueError("过滤参数必须是正整数")
        return v


class FilterGenes(BaseModel):
    """Input schema for the filter_genes preprocessing tool."""

    min_counts: Optional[int] = Field(
        default=None,
        description="Minimum number of counts required for a gene to pass filtering.",
    )

    min_cells: Optional[int] = Field(
        default=None,
        description="Minimum number of cells expressed required for a gene to pass filtering.",
    )

    max_counts: Optional[int] = Field(
        default=None,
        description="Maximum number of counts required for a gene to pass filtering.",
    )

    max_cells: Optional[int] = Field(
        default=None,
        description="Maximum number of cells expressed required for a gene to pass filtering.",
    )

    @field_validator("min_counts", "min_cells", "max_counts", "max_cells")
    def validate_positive_integers(cls, v: Optional[int]) -> Optional[int]:
        """验证整数参数为正数"""
        if v is not None and v <= 0:
            raise ValueError("must be positive_integers")
        return v


class SubsetCellParam(BaseModel):
    """Input schema for subsetting AnnData objects based on various criteria."""

    obs_key: Optional[str] = Field(
        default=None,
        description="Key in adata.obs to use for subsetting observations/cells.",
    )
    obs_min: Optional[float] = Field(
        default=None,
        description="Minimum value for the obs_key to include in the subset.",
    )
    obs_max: Optional[float] = Field(
        default=None,
        description="Maximum value for the obs_key to include in the subset.",
    )
    obs_value: Optional[Any] = Field(
        default=None,
        description="Exact value for the obs_key to include in the subset (adata.obs[obs_key] == obs_value).",
    )
    min_counts: Optional[int] = Field(
        default=None,
        description="Minimum number of counts required for a cell to pass filtering.",
    )
    min_genes: Optional[int] = Field(
        default=None,
        description="Minimum number of genes expressed required for a cell to pass filtering.",
    )
    max_counts: Optional[int] = Field(
        default=None,
        description="Maximum number of counts required for a cell to pass filtering.",
    )
    max_genes: Optional[int] = Field(
        default=None,
        description="Maximum number of genes expressed required for a cell to pass filtering.",
    )


class SubsetGeneParam(BaseModel):
    """Input schema for subsetting AnnData objects based on various criteria."""

    min_counts: Optional[int] = Field(
        default=None,
        description="Minimum number of counts required for a gene to pass filtering.",
    )
    min_cells: Optional[int] = Field(
        default=None,
        description="Minimum number of cells expressed required for a gene to pass filtering.",
    )
    max_counts: Optional[int] = Field(
        default=None,
        description="Maximum number of counts required for a gene to pass filtering.",
    )
    max_cells: Optional[int] = Field(
        default=None,
        description="Maximum number of cells expressed required for a gene to pass filtering.",
    )
    var_key: Optional[str] = Field(
        default=None,
        description="Key in adata.var to use for subsetting variables/genes.",
    )
    var_min: Optional[float] = Field(
        default=None,
        description="Minimum value for the var_key to include in the subset.",
    )
    var_max: Optional[float] = Field(
        default=None,
        description="Maximum value for the var_key to include in the subset.",
    )
    highly_variable: Optional[bool] = Field(
        default=False,
        description="If True, subset to highly variable genes. Requires 'highly_variable' column in adata.var.",
    )


class CalculateQCMetrics(BaseModel):
    """Input schema for the calculate_qc_metrics preprocessing tool."""

    expr_type: str = Field(default="counts", description="Name of kind of values in X.")

    var_type: str = Field(
        default="genes", description="The kind of thing the variables are."
    )

    qc_vars: Optional[Union[List[str], str]] = Field(
        default=[],
        description=(
            "Keys for boolean columns of .var which identify variables you could want to control for "
            "mark_var tool should be called frist when you want to calculate mt, ribo, hb, and check tool output for var columns"
        ),
    )

    percent_top: Optional[List[int]] = Field(
        default=[50, 100, 200, 500],
        description="List of ranks (where genes are ranked by expression) at which the cumulative proportion of expression will be reported as a percentage.",
    )

    layer: Optional[str] = Field(
        default=None,
        description="If provided, use adata.layers[layer] for expression values instead of adata.X",
    )

    use_raw: bool = Field(
        default=False,
        description="If True, use adata.raw.X for expression values instead of adata.X",
    )
    log1p: bool = Field(
        default=True,
        description="Set to False to skip computing log1p transformed annotations.",
    )

    @field_validator("percent_top")
    def validate_percent_top(cls, v: Optional[List[int]]) -> Optional[List[int]]:
        """验证 percent_top 中的值为正整数"""
        if v is not None:
            for rank in v:
                if not isinstance(rank, int) or rank <= 0:
                    raise ValueError("percent_top 中的所有值必须是正整数")
        return v


class Log1PParam(BaseModel):
    """Input schema for the log1p preprocessing tool."""

    base: Optional[Union[int, float]] = Field(
        default=None,
        description="Base of the logarithm. Natural logarithm is used by default.",
    )

    chunked: Optional[bool] = Field(
        default=None,
        description="Process the data matrix in chunks, which will save memory.",
    )

    chunk_size: Optional[int] = Field(
        default=None,
        description="Number of observations in the chunks to process the data in.",
    )

    layer: Optional[str] = Field(
        default=None, description="Entry of layers to transform."
    )

    obsm: Optional[str] = Field(default=None, description="Entry of obsm to transform.")

    @field_validator("chunk_size")
    def validate_chunk_size(cls, v: Optional[int]) -> Optional[int]:
        """Validate chunk_size is positive integer"""
        if v is not None and v <= 0:
            raise ValueError("chunk_size must be a positive integer")
        return v


class HighlyVariableGenesParam(BaseModel):
    """Input schema for the highly_variable_genes preprocessing tool."""

    layer: Optional[str] = Field(
        default=None,
        description="If provided, use adata.layers[layer] for expression values.",
    )

    n_top_genes: Optional[int] = Field(
        default=None,
        description="Number of highly-variable genes to keep. Mandatory if `flavor='seurat_v3'",
    )

    min_disp: Optional[float] = Field(
        default=0.5, description="Minimum dispersion cutoff for gene selection."
    )

    max_disp: Optional[float] = Field(
        default=np.inf, description="Maximum dispersion cutoff for gene selection."
    )
    min_mean: Optional[float] = Field(
        default=0.0125, description="Minimum mean expression cutoff for gene selection."
    )
    max_mean: Optional[float] = Field(
        default=3, description="Maximum mean expression cutoff for gene selection."
    )
    span: Optional[float] = Field(
        default=0.3,
        description="Fraction of data used for loess model fit in seurat_v3.",
        gt=0,
        lt=1,
    )
    n_bins: Optional[int] = Field(
        default=20, description="Number of bins for mean expression binning.", gt=0
    )
    flavor: Optional[
        Literal["seurat", "cell_ranger", "seurat_v3", "seurat_v3_paper"]
    ] = Field(
        default="seurat", description="Method for identifying highly variable genes."
    )
    subset: Optional[bool] = Field(
        default=False, description="Inplace subset to highly-variable genes if True."
    )
    batch_key: Optional[str] = Field(
        default=None, description="Key in adata.obs for batch information."
    )

    check_values: Optional[bool] = Field(
        default=True, description="Check if counts are integers for seurat_v3 flavor."
    )

    @field_validator("n_top_genes", "n_bins")
    def validate_positive_integers(cls, v: Optional[int]) -> Optional[int]:
        """Validate positive integers"""
        if v is not None and v <= 0:
            raise ValueError("must be a positive integer")
        return v

    @field_validator("span")
    def validate_span(cls, v: float) -> float:
        """Validate span is between 0 and 1"""
        if v <= 0 or v >= 1:
            raise ValueError("span must be between 0 and 1")
        return v


class RegressOutParam(BaseModel):
    """Input schema for the regress_out preprocessing tool."""

    keys: Union[str, List[str]] = Field(
        description="Keys for observation annotation on which to regress on."
    )
    layer: Optional[str] = Field(
        default=None, description="If provided, which element of layers to regress on."
    )
    n_jobs: Optional[int] = Field(
        default=None, description="Number of jobs for parallel computation.", gt=0
    )

    @field_validator("n_jobs")
    def validate_n_jobs(cls, v: Optional[int]) -> Optional[int]:
        """Validate n_jobs is positive integer"""
        if v is not None and v <= 0:
            raise ValueError("n_jobs must be a positive integer")
        return v

    @field_validator("keys")
    def validate_keys(cls, v: Union[str, List[str]]) -> Union[str, List[str]]:
        """Ensure keys is either a string or list of strings"""
        if isinstance(v, str):
            return v
        elif isinstance(v, list) and all(isinstance(item, str) for item in v):
            return v
        raise ValueError("keys must be a string or list of strings")


class ScaleParam(BaseModel):
    """Input schema for the scale preprocessing tool."""

    zero_center: bool = Field(
        default=True,
        description="If False, omit zero-centering variables to handle sparse input efficiently.",
    )

    max_value: Optional[float] = Field(
        default=None,
        description="Clip (truncate) to this value after scaling. If None, do not clip.",
    )

    layer: Optional[str] = Field(
        default=None, description="If provided, which element of layers to scale."
    )

    obsm: Optional[str] = Field(
        default=None, description="If provided, which element of obsm to scale."
    )

    mask_obs: Optional[Union[str, bool]] = Field(
        default=None,
        description="Boolean mask or string referring to obs column for subsetting observations.",
    )

    @field_validator("max_value")
    def validate_max_value(cls, v: Optional[float]) -> Optional[float]:
        """Validate max_value is positive if provided"""
        if v is not None and v <= 0:
            raise ValueError("max_value must be positive if provided")
        return v


class CombatParam(BaseModel):
    """Input schema for the combat batch effect correction tool."""

    key: str = Field(
        default="batch",
        description="Key to a categorical annotation from adata.obs that will be used for batch effect removal.",
    )

    covariates: Optional[List[str]] = Field(
        default=None,
        description="Additional covariates besides the batch variable such as adjustment variables or biological condition.",
    )

    @field_validator("key")
    def validate_key(cls, v: str) -> str:
        """Validate key is not empty"""
        if not v.strip():
            raise ValueError("key cannot be empty")
        return v

    @field_validator("covariates")
    def validate_covariates(cls, v: Optional[List[str]]) -> Optional[List[str]]:
        """Validate covariates are non-empty strings if provided"""
        if v is not None:
            if not all(isinstance(item, str) and item.strip() for item in v):
                raise ValueError("covariates must be non-empty strings")
        return v


class ScrubletParam(BaseModel):
    """Input schema for the scrublet doublet prediction tool."""

    adata_sim: Optional[str] = Field(
        default=None,
        description="Optional path to AnnData object with simulated doublets.",
    )

    batch_key: Optional[str] = Field(
        default=None, description="Key in adata.obs for batch information."
    )

    sim_doublet_ratio: float = Field(
        default=2.0,
        description="Number of doublets to simulate relative to observed transcriptomes.",
        gt=0,
    )

    expected_doublet_rate: float = Field(
        default=0.05,
        description="Estimated doublet rate for the experiment.",
        ge=0,
        le=1,
    )

    stdev_doublet_rate: float = Field(
        default=0.02,
        description="Uncertainty in the expected doublet rate.",
        ge=0,
        le=1,
    )

    synthetic_doublet_umi_subsampling: float = Field(
        default=1.0,
        description="Rate for sampling UMIs when creating synthetic doublets.",
        gt=0,
        le=1,
    )

    knn_dist_metric: str = Field(
        default="euclidean",
        description="Distance metric used when finding nearest neighbors.",
    )

    normalize_variance: bool = Field(
        default=True,
        description="Normalize data such that each gene has variance of 1.",
    )

    log_transform: bool = Field(
        default=False, description="Whether to log-transform the data prior to PCA."
    )

    mean_center: bool = Field(
        default=True, description="Center data such that each gene has mean of 0."
    )

    n_prin_comps: int = Field(
        default=30,
        description="Number of principal components used for embedding.",
        gt=0,
    )

    use_approx_neighbors: Optional[bool] = Field(
        default=None, description="Use approximate nearest neighbor method (annoy)."
    )

    get_doublet_neighbor_parents: bool = Field(
        default=False,
        description="Return parent transcriptomes that generated doublet neighbors.",
    )

    n_neighbors: Optional[int] = Field(
        default=None,
        description="Number of neighbors used to construct KNN graph.",
        gt=0,
    )

    threshold: Optional[float] = Field(
        default=None,
        description="Doublet score threshold for calling a transcriptome a doublet.",
        ge=0,
        le=1,
    )

    @field_validator(
        "sim_doublet_ratio",
        "expected_doublet_rate",
        "stdev_doublet_rate",
        "synthetic_doublet_umi_subsampling",
        "n_prin_comps",
        "n_neighbors",
    )
    def validate_positive_numbers(
        cls, v: Optional[Union[int, float]]
    ) -> Optional[Union[int, float]]:
        """Validate positive numbers where applicable"""
        if v is not None and v <= 0:
            raise ValueError("must be a positive number")
        return v

    @field_validator("knn_dist_metric")
    def validate_knn_dist_metric(cls, v: str) -> str:
        """Validate distance metric is supported"""
        valid_metrics = ["euclidean", "manhattan", "cosine", "correlation"]
        if v.lower() not in valid_metrics:
            raise ValueError(f"knn_dist_metric must be one of {valid_metrics}")
        return v.lower()


class NeighborsParam(BaseModel):
    """Input schema for the neighbors graph construction tool."""

    n_neighbors: int = Field(
        default=15,
        description="Size of local neighborhood used for manifold approximation.",
        gt=1,
        le=100,
    )

    n_pcs: Optional[int] = Field(
        default=None,
        description="Number of PCs to use. If None, automatically determined.",
        ge=0,
    )

    use_rep: Optional[str] = Field(
        default=None, description="Key for .obsm to use as representation."
    )

    knn: bool = Field(
        default=True,
        description="Whether to use hard threshold for neighbor restriction.",
    )

    method: Literal["umap", "gauss"] = Field(
        default="umap",
        description="Method for computing connectivities ('umap' or 'gauss').",
    )

    transformer: Optional[str] = Field(
        default=None,
        description="Approximate kNN search implementation ('pynndescent' or 'rapids').",
    )

    metric: str = Field(default="euclidean", description="Distance metric to use.")

    metric_kwds: Dict[str, Any] = Field(
        default_factory=dict, description="Options for the distance metric."
    )

    random_state: int = Field(default=0, description="Random seed for reproducibility.")

    key_added: Optional[str] = Field(
        default=None, description="Key prefix for storing neighbor results."
    )

    @field_validator("n_neighbors", "n_pcs")
    def validate_positive_integers(cls, v: Optional[int]) -> Optional[int]:
        """Validate positive integers where applicable"""
        if v is not None and v <= 0:
            raise ValueError("must be a positive integer")
        return v

    @field_validator("method")
    def validate_method(cls, v: str) -> str:
        """Validate method is supported"""
        if v not in ["umap", "gauss"]:
            raise ValueError("method must be either 'umap' or 'gauss'")
        return v

    @field_validator("transformer")
    def validate_transformer(cls, v: Optional[str]) -> Optional[str]:
        """Validate transformer option is supported"""
        if v is not None and v not in ["pynndescent", "rapids"]:
            raise ValueError("transformer must be either 'pynndescent' or 'rapids'")
        return v


class NormalizeTotalParam(BaseModel):
    """Input schema for the normalize_total preprocessing tool."""

    target_sum: Optional[float] = Field(
        default=None,
        description="If None, after normalization, each cell has a total count equal to the median of total counts before normalization. If a number is provided, each cell will have this total count after normalization.",
    )

    exclude_highly_expressed: bool = Field(
        default=False,
        description="Exclude highly expressed genes for the computation of the normalization factor for each cell.",
    )

    max_fraction: float = Field(
        default=0.05,
        description="If exclude_highly_expressed=True, consider cells as highly expressed that have more counts than max_fraction of the original total counts in at least one cell.",
        gt=0,
        le=1,
    )

    key_added: Optional[str] = Field(
        default=None,
        description="Name of the field in adata.obs where the normalization factor is stored.",
    )

    layer: Optional[str] = Field(
        default=None,
        description="Layer to normalize instead of X. If None, X is normalized.",
    )

    layers: Optional[Union[Literal["all"], List[str]]] = Field(
        default=None,
        description="List of layers to normalize. If 'all', normalize all layers.",
    )

    layer_norm: Optional[str] = Field(
        default=None, description="Specifies how to normalize layers."
    )

    inplace: bool = Field(
        default=True,
        description="Whether to update adata or return dictionary with normalized copies.",
    )

    @field_validator("target_sum")
    def validate_target_sum(cls, v: Optional[float]) -> Optional[float]:
        """Validate target_sum is positive if provided"""
        if v is not None and v <= 0:
            raise ValueError("target_sum must be positive")
        return v

    @field_validator("max_fraction")
    def validate_max_fraction(cls, v: float) -> float:
        """Validate max_fraction is between 0 and 1"""
        if v <= 0 or v > 1:
            raise ValueError("max_fraction must be between 0 and 1")
        return v



from __future__ import annotations

from pydantic import Field, BaseModel,ConfigDict


class AdataModel(BaseModel):
    """Input schema for the adata tool."""
    sampleid: str | None = Field(default=None, description="adata sampleid")
    adtype: str = Field(default="exp", description="The input adata.X data type for preprocess/analysis/plotting")

    model_config = ConfigDict(
        extra="ignore"
    )

class AdataInfo(BaseModel):
    """Input schema for the adata tool."""
    sampleid: str | None = Field(default=None, description="adata sampleid")
    adtype: str = Field(default="exp", description="The input adata.X data type for preprocess/analysis/plotting")

    model_config = ConfigDict(
        extra="ignore"
    )



from pathlib import Path
import scanpy as sc
from fastmcp.tools.tool import Tool
from fastmcp.exceptions import ToolError
from scmcp_shared.schema.preset import AdataInfo
from scmcp_shared.schema.preset.io import *
from scmcp_shared.util import filter_args, forward_request, get_ads
from scmcp_shared.mcp_base import BaseMCP


class ScanpyIOMCP(BaseMCP):
    def __init__(
        self,
        include_tools: list = None,
        exclude_tools: list = None,
        AdataInfo=AdataInfo,
    ):
        """Initialize ScanpyIOMCP with optional tool filtering."""
        super().__init__("SCMCP-IO-Server", include_tools, exclude_tools, AdataInfo)

    def _tool_read(self):
        def _read(request: ReadParam, adinfo: self.AdataInfo = self.AdataInfo()):
            """
            Read data from 10X directory or various file formats (h5ad, 10x, text files, etc.).
            """
            try:
                res = forward_request("io_read", request, adinfo)
                if res is not None:
                    return res
                kwargs = request.model_dump()
                file = Path(kwargs.get("filename", None))
                if file.is_dir():
                    kwargs["path"] = kwargs["filename"]
                    func_kwargs = filter_args(request, sc.read_10x_mtx)
                    adata = sc.read_10x_mtx(kwargs["path"], **func_kwargs)
                elif file.is_file():
                    func_kwargs = filter_args(request, sc.read)
                    adata = sc.read(**func_kwargs)
                    if kwargs.get("transpose", False):
                        adata = adata.T
                else:
                    raise FileNotFoundError(f"{kwargs['filename']} does not exist")

                ads = get_ads()
                if adinfo.sampleid is not None:
                    ads.active_id = adinfo.sampleid
                else:
                    ads.active_id = f"adata{len(ads.adata_dic[adinfo.adtype])}"

                adata.layers["counts"] = adata.X
                adata.var_names_make_unique()
                adata.obs_names_make_unique()
                adata.obs["scmcp_sampleid"] = adinfo.sampleid or ads.active_id
                ads.set_adata(adata, adinfo=adinfo)
                return [
                    {
                        "sampleid": adinfo.sampleid or ads.active_id,
                        "adtype": adinfo.adtype,
                        "adata": adata,
                        "adata.obs_names[:10]": adata.obs_names[:10],
                        "adata.var_names[:10]": adata.var_names[:10],
                        "notice": "check obs_names and var_names. transpose the data if needed",
                    }
                ]
            except ToolError as e:
                raise ToolError(e)
            except Exception as e:
                if hasattr(e, "__context__") and e.__context__:
                    raise ToolError(e.__context__)
                else:
                    raise ToolError(e)

        return Tool.from_function(_read, name="read", enabled=True)

    def _tool_write(self):
        def _write(request: WriteParam, adinfo: self.AdataInfo = self.AdataInfo()):
            """save adata into a file."""
            try:
                res = forward_request("io_write", request, adinfo)
                if res is not None:
                    return res
                ads = get_ads()
                adata = ads.get_adata(adinfo=adinfo)
                kwargs = request.model_dump()
                sc.write(kwargs["filename"], adata)
                return {"filename": kwargs["filename"], "msg": "success to save file"}
            except ToolError as e:
                raise ToolError(e)
            except Exception as e:
                if hasattr(e, "__context__") and e.__context__:
                    raise ToolError(e.__context__)
                else:
                    raise ToolError(e)

        return Tool.from_function(_write, name="write", enabled=True)


# Create an instance of the class
io_mcp = ScanpyIOMCP().mcp



from fastmcp import FastMCP
from ..agent import rag_agent
from pydantic import Field

rag_mcp = FastMCP("RAG-Server")


@rag_mcp.tool(tags={"rag"})
def retrieve_knowledge(
    task: str = Field(description="The tasks or questions that needs to be solved"),
):
    """search function and parameters that can be used to solve the user's tasks or questions"""
    return rag_agent(task, software="scmcp")



from typing import Optional, Union, List, Literal, Tuple, Mapping
from pydantic import Field, field_validator, BaseModel


# 创建 Mixin 类处理特定功能
class LegendMixin:
    """处理图例相关的字段"""

    legend_fontsize: Optional[Union[int, float, str]] = Field(
        default=None, description="Numeric size in pt or string describing the size."
    )

    legend_fontweight: Union[int, str] = Field(
        default="bold",
        description="Legend font weight. A numeric value in range 0-1000 or a string.",
    )

    legend_loc: str = Field(
        default="right margin",
        description="Location of legend, either 'on data', 'right margin' or a valid keyword for the loc parameter.",
    )

    legend_fontoutline: Optional[int] = Field(
        default=None, description="Line width of the legend font outline in pt."
    )


class ColorMappingMixin:
    """处理颜色映射相关的字段"""

    color_map: Optional[str] = Field(
        default=None, description="Color map to use for continuous variables."
    )

    palette: Optional[Union[str, List[str], Mapping[str, str]]] = Field(
        default=None,
        description="Colors to use for plotting categorical annotation groups.",
    )

    vmax: Optional[Union[str, float, List[Union[str, float]]]] = Field(
        default=None,
        description="The value representing the upper limit of the color scale.",
    )

    vmin: Optional[Union[str, float, List[Union[str, float]]]] = Field(
        default=None,
        description="The value representing the lower limit of the color scale.",
    )

    vcenter: Optional[Union[str, float, List[Union[str, float]]]] = Field(
        default=None,
        description="The value representing the center of the color scale.",
    )


class FigureSizeMixin:
    """处理图形大小相关的字段"""

    figsize: Optional[Tuple[float, float]] = Field(
        default=None, description="Figure size. Format is (width, height)."
    )


# 基础可视化模型，包含所有可视化工具共享的字段
class BaseVisualizationParam(
    BaseModel, LegendMixin, ColorMappingMixin, FigureSizeMixin
):
    adata: str = Field(..., description="The AnnData object variable name.")
    """基础可视化模型，包含所有可视化工具共享的字段"""
    pass


# 基础嵌入可视化模型，包含所有嵌入可视化工具共享的字段
class BaseEmbeddingParam(BaseVisualizationParam):
    """基础嵌入可视化模型，包含所有嵌入可视化工具共享的字段"""

    color: Optional[Union[str, List[str]]] = Field(
        default=None,
        description="Keys for annotations of observations/cells or variables/genes.",
    )

    gene_symbols: Optional[str] = Field(
        default=None,
        description="Column name in .var DataFrame that stores gene symbols.",
    )

    use_raw: Optional[bool] = Field(
        default=None,
        description="Use .raw attribute of adata for coloring with gene expression.",
    )

    sort_order: bool = Field(
        default=True,
        description="For continuous annotations used as color parameter, plot data points with higher values on top of others.",
    )

    edges: bool = Field(default=False, description="Show edges between nodes.")

    edges_width: float = Field(default=0.1, description="Width of edges.")

    edges_color: Union[str, List[float], List[str]] = Field(
        default="grey", description="Color of edges."
    )

    neighbors_key: Optional[str] = Field(
        default=None, description="Where to look for neighbors connectivities."
    )

    arrows: bool = Field(default=False, description="Show arrows.")

    groups: Optional[Union[str, List[str]]] = Field(
        default=None,
        description="Restrict to a few categories in categorical observation annotation.",
    )

    components: Optional[Union[str, List[str]]] = Field(
        default=None,
        description="For instance, ['1,2', '2,3']. To plot all available components use components='all'.",
    )

    dimensions: Optional[Union[Tuple[int, int], List[Tuple[int, int]]]] = Field(
        default=None,
        description="0-indexed dimensions of the embedding to plot as integers. E.g. [(0, 1), (1, 2)].",
    )

    layer: Optional[str] = Field(
        default=None,
        description="Name of the AnnData object layer that wants to be plotted.",
    )

    projection: Literal["2d", "3d"] = Field(
        default="2d", description="Projection of plot."
    )

    size: Optional[Union[float, List[float]]] = Field(
        default=None, description="Point size. If None, is automatically computed."
    )

    frameon: Optional[bool] = Field(
        default=None, description="Draw a frame around the scatter plot."
    )

    add_outline: Optional[bool] = Field(
        default=False, description="Add outline to scatter plot points."
    )

    ncols: int = Field(default=4, description="Number of columns for multiple plots.")

    marker: Union[str, List[str]] = Field(
        default=".", description="Matplotlib marker style for points."
    )


# 重构 ScatterParam 作为基础散点图模型
class BaseScatterParam(BaseVisualizationParam):
    """基础散点图模型"""

    x: Optional[str] = Field(default=None, description="x coordinate.")

    y: Optional[str] = Field(default=None, description="y coordinate.")

    color: Optional[Union[str, List[str]]] = Field(
        default=None,
        description="Keys for annotations of observations/cells or variables/genes, or a hex color specification.",
    )

    use_raw: Optional[bool] = Field(
        default=None,
        description="Whether to use raw attribute of adata. Defaults to True if .raw is present.",
    )

    layers: Optional[Union[str, List[str]]] = Field(
        default=None,
        description="Use the layers attribute of adata if present: specify the layer for x, y and color.",
    )

    basis: Optional[str] = Field(
        default=None, description="Basis to use for embedding."
    )


# 使用继承关系重构 EnhancedScatterParam
class EnhancedScatterParam(BaseScatterParam):
    """Input schema for the enhanced scatter plotting tool."""

    sort_order: bool = Field(
        default=True,
        description="For continuous annotations used as color parameter, plot data points with higher values on top of others.",
    )

    alpha: Optional[float] = Field(
        default=None, description="Alpha value for the plot.", ge=0, le=1
    )

    groups: Optional[Union[str, List[str]]] = Field(
        default=None,
        description="Restrict to a few categories in categorical observation annotation.",
    )

    components: Optional[Union[str, List[str]]] = Field(
        default=None,
        description="For instance, ['1,2', '2,3']. To plot all available components use components='all'.",
    )

    projection: Literal["2d", "3d"] = Field(
        default="2d", description="Projection of plot."
    )

    right_margin: Optional[float] = Field(
        default=None, description="Adjust the width of the right margin."
    )

    left_margin: Optional[float] = Field(
        default=None, description="Adjust the width of the left margin."
    )

    @field_validator("alpha")
    def validate_alpha(cls, v: Optional[float]) -> Optional[float]:
        """Validate alpha is between 0 and 1"""
        if v is not None and (v < 0 or v > 1):
            raise ValueError("alpha must be between 0 and 1")
        return v


# 创建基础统计可视化模型
class BaseStatPlotParam(BaseVisualizationParam):
    """基础统计可视化模型，包含统计图表共享的字段"""

    groupby: Optional[str] = Field(
        default=None, description="The key of the observation grouping to consider."
    )

    log: bool = Field(default=False, description="Plot on logarithmic axis.")

    use_raw: Optional[bool] = Field(
        default=None, description="Use raw attribute of adata if present."
    )

    var_names: Optional[Union[str, List[str]]] = Field(
        default=None,
        description="var_names should be a valid subset of adata.var_names.",
    )

    layer: Optional[str] = Field(
        default=None,
        description="Name of the AnnData object layer that wants to be plotted.",
    )

    gene_symbols: Optional[str] = Field(
        default=None,
        description="Column name in .var DataFrame that stores gene symbols.",
    )

    # 添加共享的小提琴图相关字段
    stripplot: bool = Field(
        default=True, description="Add a stripplot on top of the violin plot."
    )

    jitter: Union[float, bool] = Field(
        default=True,
        description="Add jitter to the stripplot (only when stripplot is True).",
    )

    size: int = Field(default=1, description="Size of the jitter points.", gt=0)

    order: Optional[List[str]] = Field(
        default=None, description="Order in which to show the categories."
    )

    scale: Literal["area", "count", "width"] = Field(
        default="width",
        description="The method used to scale the width of each violin.",
    )

    @field_validator("size")
    def validate_size(cls, v: int) -> int:
        """Validate size is positive"""
        if v <= 0:
            raise ValueError("size must be a positive integer")
        return v


# 添加缺失的 BaseMatrixParam 类
class BaseMatrixParam(BaseVisualizationParam):
    """基础矩阵可视化模型，包含所有矩阵可视化工具共享的字段"""

    var_names: Union[List[str], Mapping[str, List[str]]] = Field(
        default=None,
        description="var_names should be a valid subset of adata.var_names or a mapping where the key is used as label to group the values.",
    )
    groupby: Union[str, List[str]] = Field(
        ...,  # Required field
        description="The key of the observation grouping to consider.",
    )
    use_raw: Optional[bool] = Field(
        default=None, description="Use raw attribute of adata if present."
    )
    log: bool = Field(default=False, description="Plot on logarithmic axis.")
    dendrogram: Union[bool, str] = Field(
        default=False,
        description="If True or a valid dendrogram key, a dendrogram based on the hierarchical clustering between the groupby categories is added.",
    )

    gene_symbols: Optional[str] = Field(
        default=None,
        description="Column name in .var DataFrame that stores gene symbols.",
    )

    var_group_positions: Optional[List[Tuple[int, int]]] = Field(
        default=None,
        description="Use this parameter to highlight groups of var_names with brackets or color blocks between the given start and end positions.",
    )

    var_group_labels: Optional[List[str]] = Field(
        default=None,
        description="Labels for each of the var_group_positions that want to be highlighted.",
    )

    layer: Optional[str] = Field(
        default=None,
        description="Name of the AnnData object layer that wants to be plotted.",
    )


# 重构 HeatmapParam
class HeatmapParam(BaseMatrixParam):
    """Input schema for the heatmap plotting tool."""

    num_categories: int = Field(
        default=7,
        description="Only used if groupby observation is not categorical. This value determines the number of groups into which the groupby observation should be subdivided.",
        gt=0,
    )

    var_group_rotation: Optional[float] = Field(
        default=None,
        description="Label rotation degrees. By default, labels larger than 4 characters are rotated 90 degrees.",
    )

    standard_scale: Optional[Literal["var", "obs"]] = Field(
        default=None,
        description="Whether or not to standardize that dimension between 0 and 1.",
    )

    swap_axes: bool = Field(
        default=False,
        description="By default, the x axis contains var_names and the y axis the groupby categories. By setting swap_axes then x are the groupby categories and y the var_names.",
    )

    show_gene_labels: Optional[bool] = Field(
        default=None,
        description="By default gene labels are shown when there are 50 or less genes. Otherwise the labels are removed.",
    )

    @field_validator("num_categories")
    def validate_num_categories(cls, v: int) -> int:
        """Validate num_categories is positive"""
        if v <= 0:
            raise ValueError("num_categories must be a positive integer")
        return v


# 重构 TracksplotParam
class TracksplotParam(BaseMatrixParam):
    """Input schema for the tracksplot plotting tool."""

    # 所有需要的字段已经在 BaseMatrixParam 中定义


# 重构 ViolinParam
class ViolinParam(BaseStatPlotParam):
    """Input schema for the violin plotting tool."""

    keys: Union[str, List[str]] = Field(
        ...,  # Required field
        description="Keys for accessing variables of adata.var or adata.obs. or variables of adata.obsm when obsm_key is not None.",
    )
    use_obsm: str = Field(
        default=None, description="using data of adata.obsm instead of adata.X"
    )
    stripplot: bool = Field(
        default=True, description="Add a stripplot on top of the violin plot."
    )
    jitter: Union[float, bool] = Field(
        default=True,
        description="Add jitter to the stripplot (only when stripplot is True).",
    )
    size: int = Field(default=1, description="Size of the jitter points.", gt=0)
    scale: Literal["area", "count", "width"] = Field(
        default="width",
        description="The method used to scale the width of each violin.",
    )
    order: Optional[List[str]] = Field(
        default=None, description="Order in which to show the categories."
    )
    multi_panel: Optional[bool] = Field(
        default=None,
        description="Display keys in multiple panels also when groupby is not None.",
    )
    xlabel: str = Field(
        default="",
        description="Label of the x axis. Defaults to groupby if rotation is None, otherwise, no label is shown.",
    )
    ylabel: Optional[Union[str, List[str]]] = Field(
        default=None, description="Label of the y axis."
    )

    rotation: Optional[float] = Field(
        default=None, description="Rotation of xtick labels."
    )

    @field_validator("size")
    def validate_size(cls, v: int) -> int:
        """Validate size is positive"""
        if v <= 0:
            raise ValueError("size must be a positive integer")
        return v


# 重构 MatrixplotParam
class MatrixplotParam(BaseMatrixParam):
    """Input schema for the matrixplot plotting tool."""

    num_categories: int = Field(
        default=7,
        description="Only used if groupby observation is not categorical. This value determines the number of groups into which the groupby observation should be subdivided.",
        gt=0,
    )

    cmap: Optional[str] = Field(
        default="viridis", description="String denoting matplotlib color map."
    )

    colorbar_title: Optional[str] = Field(
        default="Mean expression\nin group",
        description="Title for the color bar. New line character (\\n) can be used.",
    )

    var_group_rotation: Optional[float] = Field(
        default=None,
        description="Label rotation degrees. By default, labels larger than 4 characters are rotated 90 degrees.",
    )

    standard_scale: Optional[Literal["var", "group"]] = Field(
        default=None,
        description="Whether or not to standardize the given dimension between 0 and 1.",
    )

    swap_axes: bool = Field(
        default=False,
        description="By default, the x axis contains var_names and the y axis the groupby categories. By setting swap_axes then x are the groupby categories and y the var_names.",
    )
    use_obsm: str = Field(
        default=None, description="using data of adata.obsm instead of adata.X"
    )

    @field_validator("num_categories")
    def validate_num_categories(cls, v: int) -> int:
        """Validate num_categories is positive"""
        if v <= 0:
            raise ValueError("num_categories must be a positive integer")
        return v


# 重构 DotplotParam
class DotplotParam(BaseMatrixParam):
    """Input schema for the dotplot plotting tool."""

    expression_cutoff: float = Field(
        default=0.0,
        description="Expression cutoff that is used for binarizing the gene expression.",
    )

    mean_only_expressed: bool = Field(
        default=False,
        description="If True, gene expression is averaged only over the cells expressing the given genes.",
    )

    standard_scale: Optional[Literal["var", "group"]] = Field(
        default=None,
        description="Whether or not to standardize that dimension between 0 and 1.",
    )

    swap_axes: bool = Field(
        default=False,
        description="By default, the x axis contains var_names and the y axis the groupby categories. By setting swap_axes then x are the groupby categories and y the var_names.",
    )

    dot_max: Optional[float] = Field(
        default=None, description="The maximum size of the dots."
    )

    dot_min: Optional[float] = Field(
        default=None, description="The minimum size of the dots."
    )

    smallest_dot: Optional[float] = Field(
        default=None, description="The smallest dot size."
    )
    var_group_rotation: Optional[float] = Field(
        default=None,
        description="Label rotation degrees. By default, labels larger than 4 characters are rotated 90 degrees.",
    )

    colorbar_title: Optional[str] = Field(
        default="Mean expression\nin group",
        description="Title for the color bar. New line character (\\n) can be used.",
    )

    size_title: Optional[str] = Field(
        default="Fraction of cells\nin group (%)",
        description="Title for the size legend. New line character (\\n) can be used.",
    )


# 重构 RankGenesGroupsParam
class RankGenesGroupsParam(BaseVisualizationParam):
    """Input schema for the rank_genes_groups plotting tool."""

    n_genes: int = Field(default=20, description="Number of genes to show.", gt=0)

    gene_symbols: Optional[str] = Field(
        default=None,
        description="Column name in `.var` DataFrame that stores gene symbols.",
    )

    groupby: Optional[str] = Field(
        default=None, description="The key of the observation grouping to consider."
    )

    groups: Optional[Union[str, List[str]]] = Field(
        default=None, description="Subset of groups, e.g. ['g1', 'g2', 'g3']."
    )

    key: Optional[str] = Field(
        default="rank_genes_groups",
        description="Key used to store the rank_genes_groups parameters.",
    )

    fontsize: int = Field(default=8, description="Fontsize for gene names.")

    ncols: int = Field(default=4, description="Number of columns.")

    sharey: bool = Field(
        default=True,
        description="Controls if the y-axis of each panels should be shared.",
    )

    @field_validator("n_genes", "fontsize")
    def validate_positive_int(cls, v: int) -> int:
        """Validate positive integers"""
        if v <= 0:
            raise ValueError("Value must be a positive integer")
        return v


# 重构 ClusterMapModel
class ClusterMapParam(BaseModel):
    """Input schema for the clustermap plotting tool."""

    obs_keys: Optional[str] = Field(
        default=None,
        description="key column in adata.obs, categorical annotation to plot with a different color map.",
    )
    use_raw: Optional[bool] = Field(
        default=None,
        description="Whether to use `raw` attribute of `adata`. Defaults to `True` if `.raw` is present.",
    )


# 重构 StackedViolinParam
class StackedViolinParam(BaseStatPlotParam):
    """Input schema for the stacked_violin plotting tool."""

    stripplot: bool = Field(
        default=True, description="Add a stripplot on top of the violin plot."
    )

    jitter: Union[float, bool] = Field(
        default=True,
        description="Add jitter to the stripplot (only when stripplot is True).",
    )

    size: int = Field(default=1, description="Size of the jitter points.", gt=0)

    order: Optional[List[str]] = Field(
        default=None, description="Order in which to show the categories."
    )

    scale: Literal["area", "count", "width"] = Field(
        default="width",
        description="The method used to scale the width of each violin.",
    )

    swap_axes: bool = Field(
        default=False, description="Swap axes such that observations are on the x-axis."
    )

    @field_validator("size")
    def validate_size(cls, v: int) -> int:
        """Validate size is positive"""
        if v <= 0:
            raise ValueError("size must be a positive integer")
        return v


# 重构 TrackingParam
class TrackingParam(BaseVisualizationParam):
    """Input schema for the tracking plotting tool."""

    groupby: str = Field(
        ...,  # Required field
        description="The key of the observation grouping to consider.",
    )

    min_group_size: int = Field(
        default=1,
        description="Minimal number of cells in a group for the group to be considered.",
        gt=0,
    )

    min_split_size: int = Field(
        default=1,
        description="Minimal number of cells in a split for the split to be shown.",
        gt=0,
    )

    @field_validator("min_group_size", "min_split_size")
    def validate_positive_int(cls, v: int) -> int:
        """Validate positive integers"""
        if v <= 0:
            raise ValueError("Value must be a positive integer")
        return v


# 重构 EmbeddingDensityParam
class EmbeddingDensityParam(BaseEmbeddingParam):
    """Input schema for the embedding_density plotting tool."""

    basis: str = Field(
        ...,  # Required field
        description="Basis to use for embedding.",
    )

    key: Optional[str] = Field(
        default=None,
        description="Key for annotation of observations/cells or variables/genes.",
    )

    convolve: Optional[float] = Field(
        default=None, description="Sigma for Gaussian kernel used for convolution."
    )

    alpha: float = Field(
        default=0.5, description="Alpha value for the plot.", ge=0, le=1
    )

    @field_validator("alpha")
    def validate_alpha(cls, v: float) -> float:
        """Validate alpha is between 0 and 1"""
        if v < 0 or v > 1:
            raise ValueError("alpha must be between 0 and 1")
        return v


class PCAParam(BaseEmbeddingParam):
    """Input schema for the PCA plotting tool."""

    annotate_var_explained: bool = Field(
        default=False, description="Annotate the explained variance."
    )


# 重构 UMAP 模型
class UMAPParam(BaseEmbeddingParam):
    """Input schema for the UMAP plotting tool."""

    # 所有需要的字段已经在 BaseEmbeddingParam 中定义


# 重构 TSNE 模型
class TSNEParam(BaseEmbeddingParam):
    """Input schema for the TSNE plotting tool."""

    # 所有需要的字段已经在 BaseEmbeddingParam 中定义


# 重构 DiffusionMapParam
class DiffusionMapParam(BaseEmbeddingParam):
    """Input schema for the diffusion map plotting tool."""

    # 所有需要的字段已经在 BaseEmbeddingParam 中定义


class HighestExprGenesParam(BaseVisualizationParam):
    """Input schema for the highest_expr_genes plotting tool."""

    n_top: int = Field(default=30, description="Number of top genes to plot.", gt=0)

    gene_symbols: Optional[str] = Field(
        default=None,
        description="Key for field in .var that stores gene symbols if you do not want to use .var_names.",
    )

    log: bool = Field(default=False, description="Plot x-axis in log scale.")

    @field_validator("n_top")
    def validate_n_top(cls, v: int) -> int:
        """Validate n_top is positive"""
        if v <= 0:
            raise ValueError("n_top must be a positive integer")
        return v


class HighlyVariableGenesParam(BaseVisualizationParam):
    """Input schema for the highly_variable_genes plotting tool."""

    log: bool = Field(default=False, description="Plot on logarithmic axes.")

    highly_variable_genes: bool = Field(
        default=True, description="Whether to plot highly variable genes or all genes."
    )


class PCAVarianceRatioParam(BaseVisualizationParam):
    """Input schema for the pca_variance_ratio plotting tool."""

    n_pcs: int = Field(default=30, description="Number of PCs to show.", gt=0)

    log: bool = Field(default=False, description="Plot on logarithmic scale.")

    @field_validator("n_pcs")
    def validate_n_pcs(cls, v: int) -> int:
        """Validate n_pcs is positive"""
        if v <= 0:
            raise ValueError("n_pcs must be a positive integer")
        return v


# ... existing code ...


class RankGenesGroupsDotplotParam(BaseMatrixParam):
    """Input schema for the rank_genes_groups_dotplot plotting tool."""

    groups: Optional[Union[str, List[str]]] = Field(
        default=None, description="The groups for which to show the gene ranking."
    )
    n_genes: Optional[int] = Field(
        default=None,
        description="Number of genes to show. This can be a negative number to show down regulated genes. Ignored if var_names is passed.",
    )
    values_to_plot: Optional[
        Literal[
            "scores",
            "logfoldchanges",
            "pvals",
            "pvals_adj",
            "log10_pvals",
            "log10_pvals_adj",
        ]
    ] = Field(
        default=None,
        description="Instead of the mean gene value, plot the values computed by sc.rank_genes_groups.",
    )
    min_logfoldchange: Optional[float] = Field(
        default=None,
        description="Value to filter genes in groups if their logfoldchange is less than the min_logfoldchange.",
    )
    key: Optional[str] = Field(
        default=None, description="Key used to store the ranking results in adata.uns."
    )
    var_names: Union[List[str], Mapping[str, List[str]]] = Field(
        default=None,
        description="Genes to plot. Sometimes is useful to pass a specific list of var names (e.g. genes) to check their fold changes or p-values",
    )

    @field_validator("n_genes")
    def validate_n_genes(cls, v: Optional[int]) -> Optional[int]:
        """Validate n_genes"""
        # n_genes can be positive or negative, so no validation needed
        return v


class EmbeddingParam(BaseEmbeddingParam):
    """Input schema for the embedding plotting tool."""

    basis: str = Field(
        ...,  # Required field
        description="Name of the obsm basis to use.",
    )
    use_obsm: str = Field(
        default=None, description="using data of adata.obsm instead of adata.X"
    )
    mask_obs: Optional[str] = Field(
        default=None,
        description="A boolean array or a string mask expression to subset observations.",
    )

    arrows_kwds: Optional[dict] = Field(
        default=None,
        description="Passed to matplotlib's quiver function for drawing arrows.",
    )

    scale_factor: Optional[float] = Field(
        default=None, description="Scale factor for the plot."
    )

    cmap: Optional[str] = Field(
        default=None,
        description="Color map to use for continuous variables. Overrides color_map.",
    )

    na_color: str = Field(
        default="lightgray", description="Color to use for null or masked values."
    )

    na_in_legend: bool = Field(
        default=True, description="Whether to include null values in the legend."
    )

    outline_width: Tuple[float, float] = Field(
        default=(0.3, 0.05), description="Width of the outline for highlighted points."
    )

    outline_color: Tuple[str, str] = Field(
        default=("black", "white"),
        description="Color of the outline for highlighted points.",
    )

    colorbar_loc: Optional[str] = Field(
        default="right", description="Location of the colorbar."
    )

    hspace: float = Field(default=0.25, description="Height space between panels.")

    wspace: Optional[float] = Field(
        default=None, description="Width space between panels."
    )

    title: Optional[Union[str, List[str]]] = Field(
        default=None, description="Title for the plot."
    )
